# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# ===================================================
# Create Subinterface
# $(config)interface GigabitEthernet 2.500
# $(config)description OPENSTACK_NEUTRON_INTF
# $(config)encapsulation dot1Q 500
# $(config)vrf forwarding nrouter-abc-e7d4y5
# $(config)ip address 192.168.0.1 255.255.255.0
# ===================================================
CREATE_SUBINTERFACE_WITH_ID = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>description OPENSTACK_NEUTRON_INTF</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Create Subinterface
# $(config)interface GigabitEthernet 2.500
# $(config)description OPENSTACK_NEUTRON_L3FR001_INTF
# $(config)encapsulation dot1Q 500
# $(config)vrf forwarding nrouter-e7d4y5-L3FR001
# $(config)ip address 192.168.0.1 255.255.255.0
# ===================================================
CREATE_SUBINTERFACE_REGION_ID_WITH_ID = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>description OPENSTACK_NEUTRON_%s_INTF</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Create Subinterface (External. no VRF)
# $(config)interface GigabitEthernet 2.500
# $(config)description OPENSTACK_NEUTRON_L3FR001_INTF
# $(config)encapsulation dot1Q 500
# $(config)ip address 192.168.0.1 255.255.255.0
# ===================================================
CREATE_SUBINTERFACE_EXTERNAL_WITH_ID = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>description OPENSTACK_NEUTRON_EXTERNAL_INTF</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

CREATE_SUBINTERFACE_EXT_REGION_ID_WITH_ID = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>description OPENSTACK_NEUTRON_%s_INTF</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Set primary ip for interface
# $(config)interface GigabitEthernet 2.500
# $(config)ip address 192.168.0.1 255.255.255.0
# ===================================================
SET_INTERFACE_PRIMARY_IP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Remove primary ip for interface
# $(config)interface GigabitEthernet 2.500
# $(config)no ip address 192.168.0.1 255.255.255.0
# ===================================================
REMOVE_INTERFACE_PRIMARY_IP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no ip address %s %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Add secondary ip for additional subnets
# $(config)interface GigabitEthernet 2.500
# $(config)ip address 192.168.0.1 255.255.255.0 secondary
# ===================================================
SET_INTERFACE_SECONDARY_IP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>ip address %s %s secondary</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Remove secondary ip for additional subnets
# $(config)interface GigabitEthernet 2.500
# $(config)no ip address 192.168.0.1 255.255.255.0 secondary
# ===================================================
REMOVE_INTERFACE_SECONDARY_IP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no ip address %s %s secondary</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Enable HSRP on a Subinterface
# $(config)interface GigabitEthernet0/0/0.314
# $(config)vrf forwarding nrouter-e7d4y5
# $(config)standby version 2
# $(config)standby delay minimum 30 reload 60
# $(config)standby 1621 priority 10
# $(config)standby 1621 ip 10.0.3.1
# $(config)standby 1621 timers 1 3
# ===================================================
SET_INTC_ASR_HSRP = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>standby version 2</cmd>
            <cmd>standby delay minimum 30 reload 60</cmd>
            <cmd>standby %s priority %s</cmd>
            <cmd>standby %s ip %s</cmd>
            <cmd>standby %s timers 1 3</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Enable HSRP on a External Network Subinterface
# $(config)interface GigabitEthernet0/0/0.314
# $(config)standby version 2
# $(config)standby delay minimum 30 reload 60
# $(config)standby 1621 priority 10
# $(config)standby 1621 ip 10.0.3.1
# $(config)standby 1621 timers 1 3
# $(config)standby 1621 name neutron-hsrp-1621-314
# ===================================================
SET_INTC_ASR_HSRP_EXTERNAL = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>standby version 2</cmd>
            <cmd>standby delay minimum 30 reload 60</cmd>
            <cmd>standby %s priority %s</cmd>
            <cmd>standby %s ip %s</cmd>
            <cmd>standby %s timers 1 3</cmd>
            <cmd>standby %s name neutron-hsrp-%s-%s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Enable HSRP on a External Network Subinterface secondary IP
# $(config)interface GigabitEthernet0/0/0.314
# $(config)standby 1621 ip 10.0.3.1 secondary
# ===================================================
SET_INTC_ASR_SECONDARY_HSRP_EXTERNAL = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>standby %s ip %s secondary</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Disable HSRP on a External Network Subinterface secondary IP
# $(config)interface GigabitEthernet0/0/0.314
# $(config)no standby 1621 ip 10.0.3.1 secondary
# ===================================================
REMOVE_INTC_ASR_SECONDARY_HSRP_EXTERNAL = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no standby %s ip %s secondary</cmd>
        </cli-config-data>
</config>
"""

# ===========================================================================
# Set Static source translation on an interface
# Syntax: ip nat inside source static <fixed_ip> <floating_ip>
# .......vrf <vrf_name> redundancy <hsrp group name>
# eg: $(config)ip nat inside source static 192.168.0.1 121.158.0.5
#    ..........vrf nrouter-e7d4y5 redundancy neutron-hsrp-1621-314
# ==========================================================================
SET_STATIC_SRC_TRL_NO_VRF_MATCH = """
<config>
        <cli-config-data>
            <cmd>ip nat inside source static %s %s vrf %s redundancy neutron-hsrp-%s-%s</cmd> # NOQA
        </cli-config-data>
</config>
"""

# ===========================================================================
# Remove Static source translation on an interface
# Syntax: no ip nat inside source static <fixed_ip> <floating_ip>
# .......vrf <vrf_name> redundancy <hsrp group name>
# eg: $(config)no ip nat inside source static 192.168.0.1 121.158.0.5
#    ..........vrf nrouter-e7d4y5 redundancy neutron-hsrp-1621-314
# ==========================================================================
REMOVE_STATIC_SRC_TRL_NO_VRF_MATCH = """
<config>
        <cli-config-data>
            <cmd>no ip nat inside source static %s %s vrf %s redundancy neutron-hsrp-%s-%s</cmd> # NOQA
        </cli-config-data>
</config>
"""

# =============================================================================
# Set default ip route with interface
# Syntax: ip route vrf <vrf-name> 0.0.0.0 0.0.0.0 <interface> <next hop>
# eg:
# $(config)ip route vrf nrouter-e7d4y5 0.0.0.0  0.0.0.0 po10.304 10.0.100.255
# =============================================================================
# ToDo(Hareesh): Seems unused, remove commented below after testing
# DEFAULT_ROUTE_WITH_INTF_CFG = 'ip route vrf %s 0.0.0.0 0.0.0.0 %s %s'

SET_DEFAULT_ROUTE_WITH_INTF = """
<config>
        <cli-config-data>
            <cmd>ip route vrf %s 0.0.0.0 0.0.0.0 %s %s</cmd>
        </cli-config-data>
</config>
"""

# =============================================================================
# Remove default ip route
# Syntax: ip route vrf <vrf-name> 0.0.0.0 0.0.0.0 <interface> <next hop>
# eg:
#   $(config)ip route vrf nrouter-e7d4y5 0.0.0.0 0.0.0.0 po10.304 10.0.100.255
# =============================================================================
REMOVE_DEFAULT_ROUTE_WITH_INTF = """
<config>
        <cli-config-data>
            <cmd>no ip route vrf %s 0.0.0.0 0.0.0.0 %s %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Create VRF definition
# $(config)vrf definition nrouter-e7d4y5
# $(config)address-family ipv4
# $(config)exit-address-family
# $(config)address-family ipv6
# $(config)exit-address-family
# ===================================================
CREATE_VRF_DEFN = """
<config>
        <cli-config-data>
            <cmd>vrf definition %s</cmd>
            <cmd>address-family ipv4</cmd>
            <cmd>exit-address-family</cmd>
            <cmd>address-family ipv6</cmd>
            <cmd>exit-address-family</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Remove VRF definition
# $(config)no vrf definition nrouter-e7d4y5
# ===================================================
REMOVE_VRF_DEFN = """
<config>
        <cli-config-data>
            <cmd>no vrf definition %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Create V6 Subinterface (With deployement id)
# $(config)interface GigabitEthernet 2.500
# $(config)description OPENSTACK_NEUTRON-Region_XY_INTF
# $(config)encapsulation dot1Q 500
# $(config)vrf forwarding nrouter-abc-e7d4y5
# $(config)ipv6 address 2001:DB8:CAFE:A::1/64
# ===================================================
CREATE_SUBINTERFACE_V6_WITH_ID = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>description OPENSTACK_NEUTRON-%s_INTF</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>vrf forwarding %s</cmd>
            <cmd>ipv6 address %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Create Subinterface (with deployment_id)
# $(config)interface GigabitEthernet 2.500
# $(config)description OPENSTACK_NEUTRON-Region_XY_INTF
# $(config)encapsulation dot1Q 500
# $(config)ipv6 address 2001:DB8:CAFE:A::1/64
# ===================================================
CREATE_SUBINTERFACE_V6_NO_VRF_WITH_ID = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>description OPENSTACK_NEUTRON-%s_INTF</cmd>
            <cmd>encapsulation dot1Q %s</cmd>
            <cmd>ipv6 address %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Enable HSRP on a Subinterface
# $(config)interface GigabitEthernet0/0/0.314
# $(config)vrf forwarding nrouter-e7d4y5
# $(config)standby version 2
# $(config)standby 1621 ipv6 autoconfig
# $(config)standby 1621 priority 10
# $(config)standby 1621 preempt
# $(config)standby 1621 timers 1 3
# $(config)standby 1621 name neutron-hsrp-1621-314
# ===================================================
SET_INTC_ASR_HSRP_V6 = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>standby version 2</cmd>
            <cmd>standby %s ipv6 autoconfig</cmd>
            <cmd>standby %s priority %s</cmd>
            <cmd>standby %s preempt</cmd>
            <cmd>standby %s authentication OPEN</cmd>
            <cmd>standby %s timers 1 3</cmd>
            <cmd>standby %s name neutron-hsrp-%s-%s</cmd>
        </cli-config-data>
</config>
"""

# =============================================================================
# Set default ipv6 route with interface
# Syntax: ipv6 route vrf <vrf-name> ::/0 <interface> nexthop-vrf default
# eg:
# $(config)ipv6 route vrf nrouter-e7d4y5 ::/0 po10.304 nexthop-vrf default
# =============================================================================
# ToDo(Hareesh): Seems unused, remove commented below after testing
# DEFAULT_ROUTE_V6_WITH_INTF_CFG = 'ipv6 route vrf %s ::/0 %s %s'

SET_DEFAULT_ROUTE_V6_WITH_INTF = """
<config>
        <cli-config-data>
            <cmd>ipv6 route vrf %s ::/0 %s nexthop-vrf default</cmd>
        </cli-config-data>
</config>
"""

# ============================================================================
# Remove default ipv6 route
# Syntax: no ipv6 route vrf <vrf-name> ::/0 <interface> nexthop-vrf default
# eg:
# $(config)no ipv6 route vrf nrouter-e7d4y5 ::/0 po10.304 nexthop-vrf default
# ============================================================================
REMOVE_DEFAULT_ROUTE_V6_WITH_INTF = """
<config>
        <cli-config-data>
            <cmd>no ipv6 route vrf %s ::/0 %s nexthop-vrf default</cmd>
        </cli-config-data>
</config>
"""

# ===========================================================================
# Set Dynamic source translation with NAT pool
# Syntax: ip nat inside source list <acl_no> pool <pool_name>
#     ....vrf <vrf_name> overload
# eg: $(config)ip nat inside source list acl_500
#    ....pool nrouter-e7d4y5-pool vrf nrouter-e7d4y5 overload
# ==========================================================================
# ToDo(Hareesh): Seems unused, remove commented below after testing
# SNAT_POOL_CFG = "ip nat inside source list %s pool %s vrf %s overload"

SET_DYN_SRC_TRL_POOL = """
<config>
        <cli-config-data>
            <cmd>ip nat inside source list %s pool %s vrf %s overload</cmd>
        </cli-config-data>
</config>
"""

# ===========================================================================
# Remove Dynamic source translation with NAT pool
# Syntax: no ip nat inside source list <acl_no> pool <pool_name>
#     ...vrf <vrf_name> overload
# eg: $(config)no ip nat inside source list acl_500
#    ..........pool nrouter-e7d4y5-pool vrf nrouter-e7d4y5 overload
# ==========================================================================
REMOVE_DYN_SRC_TRL_POOL = """
<config>
        <cli-config-data>
            <cmd>no ip nat inside source list %s pool %s vrf %s overload</cmd>
        </cli-config-data>
</config>
"""

# ===========================================================================
# Create a NAT pool
# Syntax: ip nat pool <pool_name> <start_ip> <end_ip> netmask <netmask_value>
# eg:
# $(config)ip nat pool TEST_POOL 192.168.0.20 192.168.0.35 netmask 255.255.0.0
# ==========================================================================
CREATE_NAT_POOL = """
<config>
        <cli-config-data>
            <cmd>ip nat pool %s %s %s netmask %s</cmd>
        </cli-config-data>
</config>
"""

# ===========================================================================
# Delete a NAT pool
# Syntax:no ip nat pool <pool_name> <start_ip> <end_ip> netmask <netmask_value>
# eg:
# $(config)no ip nat pool TEST_POOL 192.168.0.20 192.168.0.35
# .........netmask 255.255.0.0
# ==========================================================================
DELETE_NAT_POOL = """
<config>
        <cli-config-data>
            <cmd>no ip nat pool %s %s %s netmask %s</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Disable HSRP preempt on an interface
# $(config)interface GigabitEthernet 2.500
# $(config)no standby 1621 preempt
# ===================================================
REMOVE_INTC_ASR_HSRP_PREEMPT = """
<config>
        <cli-config-data>
            <cmd>interface %s</cmd>
            <cmd>no standby %s preempt</cmd>
        </cli-config-data>
</config>
"""

# ===================================================
# Empty snippet (for polling netconf session status)
# ===================================================
EMPTY_SNIPPET = """
<config>
        <cli-config-data>
            <cmd>do cd</cmd>
        </cli-config-data>
</config>
"""



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_utils import netutils
import six

from neutron.api.v2 import attributes as attr
from neutron.api.v2 import resource_helper
from neutron_lib.api import converters as conv
from neutron_lib import exceptions as nexception

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import cisco_constants as constants
from networking_cisco.plugins.cisco.common import utils


NEUTRON_VERSION = bc.NEUTRON_VERSION
NEUTRON_NEWTON_VERSION = bc.NEUTRON_NEWTON_VERSION


# Hosting device and hosting device template exceptions
class HostingDeviceInvalidPortValue(nexception.InvalidInput):
    message = _("Invalid value for port %(port)s")


class HostingDeviceInUse(nexception.InUse):
    message = _("Hosting device %(id)s in use.")


class HostingDeviceMgmtPortNotFound(nexception.InUse):
    message = _("Specified management port %(id)s does not exist.")


class HostingDeviceNotFound(nexception.NotFound):
    message = _("Hosting device %(id)s does not exist")


class HostingDeviceTemplateNotFound(nexception.NotFound):
    message = _("Hosting device template %(id)s does not exist")


class HostingDeviceTemplateInUse(nexception.InUse):
    message = _("Hosting device template %(id)s in use.")


class TenantBoundNotUUIDListOrNone(nexception.NetworkNotFound):
    message = _("Attribute tenant_bound must be a list of tenant ids or None")


def convert_validate_port_value(port):
    if port is None:
        return port

    if netutils.is_valid_port(port):
        return int(port)
    else:
        raise HostingDeviceInvalidPortValue(port=port)


def convert_empty_string_to_none(value):
    if value == "":
        return None
    else:
        return value


# Hosting device belong to one of the following categories:
VM_CATEGORY = 'VM'
HARDWARE_CATEGORY = 'Hardware'
NETWORK_NODE_CATEGORY = 'Network_Node'

HOSTING_DEVICE_MANAGER_ALIAS = 'dev_mgr'
DEVICE = 'hosting_device'
DEVICES = DEVICE + 's'
DEVICE_TEMPLATE = DEVICE + '_template'
DEVICE_TEMPLATES = DEVICE_TEMPLATE + 's'

AUTO_DELETE_DEFAULT = False

# Attribute Map
RESOURCE_ATTRIBUTE_MAP = {
    DEVICES: {
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'required_by_policy': True, 'is_visible': True},
        'id': {'allow_post': True, 'allow_put': False,
               'validate': {'type:uuid_or_none': None},
               'default': None, 'is_visible': True,
               'primary_key': True},
        'template_id': {'allow_post': True, 'allow_put': False,
                        'validate': {'type:uuid': None},
                        'required_by_policy': True, 'is_visible': True},
        'credentials_id': {'allow_post': True, 'allow_put': True,
                           'default': None,
                           'validate': {'type:uuid_or_none': None},
                           'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string_or_none': None},
                 'default': None, 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string_or_none': None},
                        'default': None, 'is_visible': True},
        'device_id': {'allow_post': True, 'allow_put': True,
                      'validate': {'type:string_or_none': None},
                      'default': None, 'is_visible': True},
        'admin_state_up': {'allow_post': True, 'allow_put': True,
                           'convert_to': conv.convert_to_boolean,
                           'default': True, 'is_visible': True},
        'management_ip_address': {'allow_post': True, 'allow_put': False,
                                  'validate': {'type:ip_address': None},
                                  'is_visible': True},
        'management_port_id': {'allow_post': True, 'allow_put': False,
                               'validate': {'type:uuid_or_none': None},
                               'default': None, 'is_visible': True},
        'protocol_port': {'allow_post': True, 'allow_put': False,
                          'convert_to': convert_validate_port_value,
                          'default': None, 'is_visible': True},
        'cfg_agent_id': {'allow_post': True, 'allow_put': False,
                         'validate': {'type:uuid_or_none': None},
                         'default': None, 'is_visible': True},
        'created_at': {'allow_post': False, 'allow_put': False,
                       'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'default': None, 'is_visible': True},
        'tenant_bound': {'allow_post': True, 'allow_put': True,
                         'convert_to': convert_empty_string_to_none,
                         'validate': {'type:uuid_or_none': None},
                         'default': None, 'is_visible': True},
        'auto_delete': {'allow_post': True, 'allow_put': True,
                        'convert_to': conv.convert_to_boolean,
                        'default': AUTO_DELETE_DEFAULT, 'is_visible': True},
    },
    DEVICE_TEMPLATES: {
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'required_by_policy': True, 'is_visible': True},
        'id': {'allow_post': True, 'allow_put': False,
               'validate': {'type:uuid_or_none': None}, 'default': None,
               'is_visible': True, 'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string_or_none': None},
                 'default': None, 'is_visible': True},
        'enabled': {'allow_post': True, 'allow_put': True,
                    'convert_to': conv.convert_to_boolean,
                    'default': True, 'is_visible': True},
        'host_category': {'allow_post': True, 'allow_put': False,
                          'validate': {'type:values': [VM_CATEGORY,
                                                       HARDWARE_CATEGORY,
                                                       NETWORK_NODE_CATEGORY]},
                          'required_by_policy': True, 'is_visible': True},
        #TODO(bobmel): validate service_types
        'service_types': {'allow_post': True, 'allow_put': True,
                          'default': None, 'is_visible': True},
        'image': {'allow_post': True, 'allow_put': True,
                  'validate': {'type:string_or_none': None},
                  'default': None, 'is_visible': True},
        'flavor': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:string_or_none': None},
                   'default': None, 'is_visible': True},
        'default_credentials_id': {'allow_post': True, 'allow_put': True,
                                   'validate': {'type:uuid': None},
                                   'is_visible': True},
        'configuration_mechanism': {'allow_post': True, 'allow_put': True,
                                    'validate': {'type:string_or_none': None},
                                    'is_visible': True},
        'protocol_port': {'allow_post': True, 'allow_put': True,
                          'convert_to': convert_validate_port_value,
                          'default': None, 'is_visible': True},
        'booting_time': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:non_negative': 0},
                         'convert_to': conv.convert_to_int,
                         'default': None, 'is_visible': True},
        'slot_capacity': {'allow_post': True, 'allow_put': False,
                          'validate': {'type:non_negative': 0},
                          'convert_to': conv.convert_to_int,
                          'default': 0, 'is_visible': True},
        'desired_slots_free': {'allow_post': True, 'allow_put': False,
                               'validate': {'type:non_negative': 0},
                               'convert_to': conv.convert_to_int,
                               'default': 0, 'is_visible': True},
        'tenant_bound': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:uuid_list': []},
                         'default': [], 'is_visible': True},
        'device_driver': {'allow_post': True, 'allow_put': False,
                          'convert_to': utils.convert_validate_driver_class,
                          'is_visible': True},
        'plugging_driver': {'allow_post': True, 'allow_put': False,
                            'convert_to': utils.convert_validate_driver_class,
                            'is_visible': True},
    }
}


class Ciscohostingdevicemanager(bc.extensions.ExtensionDescriptor):
    """Hosting device and template extension."""

    @classmethod
    def get_name(cls):
        return "Cisco hosting device manager"

    @classmethod
    def get_alias(cls):
        return HOSTING_DEVICE_MANAGER_ALIAS

    @classmethod
    def get_description(cls):
        return "Extension for manager of hosting devices and their templates"

    @classmethod
    def get_namespace(cls):
        # todo
        return ("http://docs.openstack.org/ext/" +
                HOSTING_DEVICE_MANAGER_ALIAS + "/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2014-03-31T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        if NEUTRON_VERSION.version[0] <= NEUTRON_NEWTON_VERSION.version[0]:
            attr.PLURALS.update(plural_mappings)
        action_map = {DEVICE: {'get_hosting_device_config': 'GET'}}
        return resource_helper.build_resource_info(plural_mappings,
                                                   RESOURCE_ATTRIBUTE_MAP,
                                                   constants.DEVICE_MANAGER,
                                                   action_map=action_map)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


@six.add_metaclass(abc.ABCMeta)
class CiscoHostingDevicePluginBase(bc.service_base.ServicePluginBase):

    def get_plugin_name(self):
        return constants.DEVICE_MANAGER

    def get_plugin_type(self):
        return constants.DEVICE_MANAGER

    def get_plugin_description(self):
        return ("Cisco Device Manager Service Plugin for management of "
                "hosting devices and their templates.")

    @abc.abstractmethod
    def create_hosting_device(self, context, hosting_device):
        pass

    @abc.abstractmethod
    def update_hosting_device(self, context, id, hosting_device):
        pass

    @abc.abstractmethod
    def delete_hosting_device(self, context, id):
        pass

    @abc.abstractmethod
    def get_hosting_device(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def get_hosting_devices(self, context, filters=None, fields=None,
                            sorts=None, limit=None, marker=None,
                            page_reverse=False):
        pass

    @abc.abstractmethod
    def get_hosting_device_config(self, context, id):
        pass

    @abc.abstractmethod
    def create_hosting_device_template(self, context,
                                       hosting_device_template):
        pass

    @abc.abstractmethod
    def update_hosting_device_template(self, context, id,
                                       hosting_device_template):
        pass

    @abc.abstractmethod
    def delete_hosting_device_template(self, context, id):
        pass

    @abc.abstractmethod
    def get_hosting_device_template(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def get_hosting_device_templates(self, context, filters=None, fields=None,
                                     sorts=None, limit=None, marker=None,
                                     page_reverse=False):
        pass



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import uuidutils
from sqlalchemy import exc as sql_exc
from sqlalchemy.orm import exc

from networking_cisco.plugins.cisco.db.l3 import l3_models
import networking_cisco.plugins.cisco.extensions.routertype as routertype

LOG = logging.getLogger(__name__)


class RoutertypeDbMixin(routertype.RoutertypePluginBase):
    """Mixin class for Router types."""

    def create_routertype(self, context, routertype):
        """Creates a router type.

        Also binds it to the specified hosting device template.
        """
        LOG.debug("create_routertype() called. Contents %s", routertype)
        rt = routertype['routertype']
        with context.session.begin(subtransactions=True):
            routertype_db = l3_models.RouterType(
                id=self._get_id(rt),
                tenant_id=rt['tenant_id'],
                name=rt['name'],
                description=rt['description'],
                template_id=rt['template_id'],
                ha_enabled_by_default=rt['ha_enabled_by_default'],
                shared=rt['shared'],
                slot_need=rt['slot_need'],
                scheduler=rt['scheduler'],
                driver=rt['driver'],
                cfg_agent_service_helper=rt['cfg_agent_service_helper'],
                cfg_agent_driver=rt['cfg_agent_driver'])
            context.session.add(routertype_db)
        return self._make_routertype_dict(routertype_db)

    def update_routertype(self, context, id, routertype):
        LOG.debug("update_routertype() called")
        rt = routertype['routertype']
        with context.session.begin(subtransactions=True):
            rt_query = context.session.query(l3_models.RouterType)
            if not rt_query.filter_by(id=id).update(rt):
                raise routertype.RouterTypeNotFound(id=id)
        return self.get_routertype(context, id)

    def delete_routertype(self, context, id):
        LOG.debug("delete_routertype() called")
        try:
            with context.session.begin(subtransactions=True):
                routertype_query = context.session.query(l3_models.RouterType)
                if not routertype_query.filter_by(id=id).delete():
                    raise routertype.RouterTypeNotFound(id=id)
        except db_exc.DBError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if isinstance(e.inner_exception, sql_exc.IntegrityError):
                    ctxt.reraise = False
                    raise routertype.RouterTypeInUse(id=id)

    def get_routertype(self, context, id, fields=None):
        LOG.debug("get_routertype() called")
        rt_db = self._get_routertype(context, id)
        return self._make_routertype_dict(rt_db, fields)

    def get_routertypes(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        LOG.debug("get_routertypes() called")
        return self._get_collection(context, l3_models.RouterType,
                                    self._make_routertype_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker,
                                    page_reverse=page_reverse)

    def get_routertype_by_id_name(self, context, id_or_name):
        return self._make_routertype_dict(
            self.get_routertype_db_by_id_name(context, id_or_name))

    def get_routertype_db_by_id_name(self, context, id_or_name):
        query = context.session.query(l3_models.RouterType)
        query = query.filter(l3_models.RouterType.id == id_or_name)
        try:
            return query.one()
        except exc.MultipleResultsFound:
            with excutils.save_and_reraise_exception():
                LOG.error('Database inconsistency: Multiple router types '
                          'with same id %s', id_or_name)
                raise routertype.RouterTypeNotFound(router_type=id_or_name)
        except exc.NoResultFound:
            query = context.session.query(l3_models.RouterType)
            query = query.filter(l3_models.RouterType.name == id_or_name)
            try:
                return query.one()
            except exc.MultipleResultsFound:
                with excutils.save_and_reraise_exception():
                    LOG.debug('Multiple router types with name %s found. '
                              'Id must be specified to allow arbitration.',
                              id_or_name)
                    raise routertype.MultipleRouterTypes(name=id_or_name)
            except exc.NoResultFound:
                with excutils.save_and_reraise_exception():
                    LOG.error('No router type with name %s found.',
                              id_or_name)
                    raise routertype.RouterTypeNotFound(id=id_or_name)

    def _get_routertype(self, context, id):
        try:
            return self._get_by_id(context, l3_models.RouterType, id)
        except exc.NoResultFound:
            raise routertype.RouterTypeNotFound(id=id)

    def _make_routertype_dict(self, routertype, fields=None):
        res = {'id': routertype['id'],
               'tenant_id': routertype['tenant_id'],
               'name': routertype['name'],
               'description': routertype['description'],
               'template_id': routertype['template_id'],
               'ha_enabled_by_default': routertype['ha_enabled_by_default'],
               'shared': routertype['shared'],
               'slot_need': routertype['slot_need'],
               'scheduler': routertype['scheduler'],
               'driver': routertype['driver'],
               'cfg_agent_service_helper': routertype[
                   'cfg_agent_service_helper'],
               'cfg_agent_driver': routertype['cfg_agent_driver']}
        return self._fields(res, fields)

    def _get_id(self, res):
        uuid = res.get('id')
        if uuid:
            return uuid
        return uuidutils.generate_uuid()



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import time

from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_exceptions as dexc
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import utils

LOG = logging.getLogger(__name__)


class DfaFailureRecovery(object):

    """Failure recovery class."""

    def __init__(self, cfg):
        super(DfaFailureRecovery, self).__init__(cfg)
        self._cfg = cfg

    @property
    def cfg(self):
        return self._cfg

    def add_events(self, **kwargs):
        """Add failure event into the queue."""

        event_q = kwargs.get('event_queue')
        pri = kwargs.get('priority')
        if not event_q or not pri:
            return

        try:
            event_type = 'server.failure.recovery'
            payload = {}
            timestamp = time.ctime()
            data = (event_type, payload)
            event_q.put((pri, timestamp, data))
            LOG.debug('Added failure recovery event to the queue.')
        except Exception as exc:
            LOG.exception('Error: %(exc)s for event %(event)s',
                          {'exc': str(exc), 'event': event_type})
            raise exc

    def failure_recovery(self, fail_info):
        """Failure recovery task.

        In case of failure in projects, network and VM create/delete, this
        task goes through all failure cases and try the request.
        """
        # Read failed entries from project database and send request
        # (create/delete - depends on failure type) to DCNM

        # 1. Try failure recovery for create project.
        LOG.info("Started failure_recovery.")
        projs = self.get_fialed_projects_entries(constants.CREATE_FAIL)
        for proj in projs:
            LOG.debug("Failure recovery for project %(name)s.", (
                {'name': proj.name}))
            # Try to create the project in DCNM
            try:
                self.dcnm_client.create_project(self.cfg.dcnm.orchestrator_id,
                                                proj.name,
                                                self.cfg.dcnm.
                                                default_partition_name,
                                                proj.dci_id)
            except dexc.DfaClientRequestFailed as e:
                LOG.error("failure_recovery: Failed to create %(proj)s "
                          "on DCNM : %(reason)s",
                          {'proj': proj.name, 'reason': str(e)})
            else:
                # Request is sent successfully, update the database.
                self.update_project_info_cache(proj.id, dci_id=proj.dci_id,
                                               name=proj.name,
                                               opcode='update')
                LOG.debug('Success on failure recovery for '
                          'project %(name)s', {'name': proj.name})

        # 1.1 Try failure recovery for update project.
        projs = self.get_fialed_projects_entries(constants.UPDATE_FAIL)
        for proj in projs:
            LOG.debug("Failure recovery for project %(name)s.", (
                {'name': proj.name}))
            # This was failure of updating DCI id of the project in DCNM.
            try:
                self.dcnm_client.update_project(proj.name,
                                                self.cfg.dcnm.
                                                default_partition_name,
                                                proj.dci_id)
            except dexc.DfaClientRequestFailed as exc:
                LOG.error("failure_recovery: Failed to update %(proj)s "
                          "on DCNM : %(reason)s",
                          {'proj': proj.name, 'reason': str(exc)})
            else:
                # Request is sent successfully, update the database.
                self.update_project_info_cache(proj.id,
                                               dci_id=proj.dci_id,
                                               name=proj.name,
                                               opcode='update')
                LOG.debug('Success on failure recovery update for '
                          'project %(name)s', {'name': proj.name})

        # 2. Try failure recovery for create network.
        nets = self.get_all_networks()
        for net in nets:
            if (net.result == constants.CREATE_FAIL
                    and net.source.lower() == 'openstack'):
                net_id = net.network_id
                try:
                    subnets = self.neutron_event.nclient.list_subnets(
                        network_id=net_id).get('subnets')
                except dexc.ConnectionFailed:
                    LOG.exception('Failed to get subnets list.')
                    continue

                for subnet in subnets:
                    tenant_name = self.get_project_name(subnet['tenant_id'])
                    snet = utils.Dict2Obj(subnet)
                    try:
                        # Check if config_profile is not NULL.
                        if not net.config_profile:
                            cfgp, fwd_mod = (
                                self.dcnm_client.
                                get_config_profile_for_network(net.name))
                            net.config_profile = cfgp
                            net.fwd_mod = fwd_mod
                        self.dcnm_client.create_network(tenant_name, net, snet,
                                                        self.dcnm_dhcp)
                    except dexc.DfaClientRequestFailed:
                        # Still is failure, only log the error.
                        LOG.error('Failed to create network %(net)s.',
                                  {'net': net.name})
                    else:
                        # Request is sent to DCNM, update the database
                        params = dict(
                            columns=dict(config_profile=net.config_profile,
                                         fwd_mod=net.fwd_mod,
                                         result=constants.RESULT_SUCCESS))
                        self.update_network(net_id, **params)
                        LOG.debug("Success on failure recovery to create "
                                  "%(net)s", {'net': net.name})

        # 3. Try Failure recovery for VM create and delete.
        instances = self.get_vms()
        for vm in instances:
            vm_info = dict(status=vm.status,
                           vm_mac=vm.mac,
                           segmentation_id=vm.segmentation_id,
                           host=vm.host,
                           port_uuid=vm.port_id,
                           net_uuid=vm.network_id,
                           oui=dict(ip_addr=vm.ip,
                                    vm_name=vm.name,
                                    vm_uuid=vm.instance_id,
                                    gw_mac=vm.gw_mac,
                                    fwd_mod=vm.fwd_mod,
                                    oui_id='cisco'))
            if vm.result == constants.CREATE_FAIL:
                try:
                    self.neutron_event.send_vm_info(str(vm.host), str(vm_info))
                except Exception as e:
                    # Failed to send info to the agent. Keep the data in the
                    # database as failure to send it later.
                    LOG.error('Failed to send VM info to agent. '
                              'Reason %s', str(e))
                else:
                    params = dict(columns=dict(
                        result=constants.RESULT_SUCCESS))
                    self.update_vm_db(vm.port_id, **params)
                    LOG.info('Created VM %(vm)s.', {'vm': vm.name})

        for vm in instances:
            if vm.result == constants.DELETE_FAIL:
                vm_info['status'] = 'down'
                try:
                    self.neutron_event.send_vm_info(str(vm.host), str(vm_info))
                except Exception as e:
                    LOG.error('Failed to send VM info to agent. '
                              'Reason %s', str(e))
                else:
                    self.delete_vm_db(vm.port_id)
                    LOG.info('Deleted VM %(vm)s from DB.',
                             {'vm': vm.name})

        # 4. Try failure recovery for delete network.
        for net in nets:
            if (net.result == constants.DELETE_FAIL
                    and net.source.lower() == 'openstack'):
                net_id = net.network_id
                segid = net.segmentation_id
                tenant_name = self.get_project_name(net.tenant_id)
                try:
                    self.dcnm_client.delete_network(tenant_name, net)
                except dexc.DfaClientRequestFailed:
                    # Still is failure, only log the error.
                    LOG.error('Failed to delete network %(net)s.',
                              {'net': net.name})
                else:
                    # Request is sent to DCNM, delete the entry
                    # from database and return the segmentation id to the
                    # pool.
                    self.delete_network_db(net_id)
                    self.segmentation_pool.add(segid)
                    LOG.debug("Success on failure recovery to deleted "
                              "%(net)s", {'net': net.name})

        # 5. Try failure recovery for delete project.
        projs = self.get_fialed_projects_entries(constants.DELETE_FAIL)
        for proj in projs:
            LOG.debug("Failure recovery for project %(name)s.", (
                {'name': proj.name}))
            # Try to delete the project in DCNM
            try:
                self.dcnm_client.delete_project(proj.name,
                                                self.cfg.dcnm.
                                                default_partition_name)
            except dexc.DfaClientRequestFailed as e:
                # Failed to delete project in DCNM.
                # Save the info and mark it as failure and retry it later.
                LOG.error("Failure recovery is failed to delete "
                          "%(project)s on DCNM : %(reason)s",
                          {'project': proj.name, 'reason': str(e)})
            else:
                # Delete was successful, now update the database.
                self.update_project_info_cache(proj.id, opcode='delete')
                LOG.debug("Success on failure recovery to deleted "
                          "%(project)s", {'project': proj.name})

        # 6. Do failure recovery for Firewall service
        self.fw_retry_failures()

        # 7. DHCP port consistency check for HA.
        if self.need_dhcp_check():
            nets = self.get_all_networks()
            for net in nets:
                net_id = net.network_id
                LOG.debug("dhcp consistency check for net id %s", net_id)
                self.correct_dhcp_ports(net_id)
            self.decrement_dhcp_check()
        LOG.info("Finished failure_recovery.")



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from networking_cisco.ml2_drivers.nexus import (
    config as nexus_config)
from networking_cisco.ml2_drivers.nexus import (
    type_nexus_vxlan as nexus_vxlan_config)
from networking_cisco.ml2_drivers.ucsm import (
    config as ucsm_config)


def list_nexus_conf_opts():
    return [
        ('ml2_cisco', nexus_config.ml2_cisco_opts),
        ('ml2_mech_cisco_nexus:<ip_address>', nexus_config.nexus_sub_opts)
    ]


def list_nexus_vxlan_type_driver_conf_opts():
    return [
        ('ml2_type_nexus_vxlan', nexus_vxlan_config.nexus_vxlan_opts),
    ]


def list_ucsm_conf_opts():
    main_group = list(ucsm_config.ml2_cisco_ucsm_opts)
    main_group.extend(ucsm_config.ml2_cisco_ucsm_common)
    return [
        ('ml2_cisco_ucsm', main_group),
        ('ml2_cisco_ucsm_ip:<ip_address>', ucsm_config.ml2_cisco_ucsm_common),
        ('sriov_multivlan_trunk', ucsm_config.sriov_opts)
    ]



# Copyright 2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.server.services.firewall.native import (
    fabric_setup_base as FP)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    asa_rest as asa)
from networking_cisco.apps.saf.server.services.firewall.native.drivers import (
    base)

LOG = logging.getLogger(__name__)


class PhyAsa(base.BaseDriver, FP.FabricApi):

    """Physical ASA Driver. """

    def __init__(self):
        LOG.info("Initializing physical ASA")
        super(PhyAsa, self).__init__()

    def initialize(self, cfg_dict):
        self.mgmt_ip_addr = cfg_dict.get('mgmt_ip_addr').strip()
        self.user = cfg_dict.get('user').strip()
        self.pwd = cfg_dict.get('pwd').strip()
        self.interface_in = cfg_dict.get('interface_in').strip()
        self.interface_out = cfg_dict.get('interface_out').strip()
        LOG.info("ASA with mgmt %s getting initialized",
                 self.mgmt_ip_addr)
        self.asa5585 = asa.Asa5585(self.mgmt_ip_addr, self.user, self.pwd)

    def populate_event_que(self, que_obj):
        LOG.info("Populate Event for PhyAsa")

    def populate_dcnm_obj(self, dcnm_obj):
        LOG.info("Populate Event for DCNM obj")

    def network_create_notif(self, tenant_id, tenant_name, cidr):
        """Network Create Notification. """
        LOG.info("Nwk Create Notif PhyAsa")

    def network_delete_notif(self, tenant_id, tenant_name, network_id):
        """Network Delete Notification. """
        LOG.info("Nwk Delete Notif PhyAsa")

    def is_device_virtual(self):
        return False

    def get_name(self):
        return 'phy_asa'

    def get_max_quota(self):
        return self.asa5585.get_quota()

    def create_fw(self, tenant_id, data):
        LOG.info("In creating phy ASA FW data is %s", data)
        tenant_name = data.get('tenant_name')
        in_ip_dict = self.get_in_ip_addr(tenant_id)
        in_gw = in_ip_dict.get('gateway')
        in_sec_gw = in_ip_dict.get('sec_gateway')
        in_serv_node = self.get_in_srvc_node_ip_addr(tenant_id)
        out_ip_dict = self.get_out_ip_addr(tenant_id)
        out_ip_gw = out_ip_dict.get('gateway')
        out_sec_gw = out_ip_dict.get('sec_gateway')
        out_serv_node = self.get_out_srvc_node_ip_addr(tenant_id)
        in_seg, in_vlan = self.get_in_seg_vlan(tenant_id)
        out_seg, out_vlan = self.get_out_seg_vlan(tenant_id)

        kw = {'params': {'tenant_name': tenant_name,
                         'in_vlan': in_vlan, 'out_vlan': out_vlan,
                         'in_ip': in_serv_node, 'in_mask': '255.255.255.0',
                         'in_gw': in_gw, 'in_sec_gw': in_sec_gw,
                         'out_ip': out_serv_node, 'out_mask': '255.255.255.0',
                         'out_gw': out_ip_gw, 'out_sec_gw': out_sec_gw,
                         'intf_in': self.interface_in,
                         'intf_out': self.interface_out}}
        status = self.asa5585.setup(**kw)
        if status is False:
            LOG.error("Physical FW instance creation failure for "
                      "tenant %s", tenant_name)
            return False

        status = self.asa5585.apply_policy(data)
        if status is False:
            LOG.error("Applying FW policy failure for tenant %s",
                      tenant_name)

        return status

    def delete_fw(self, tenant_id, data):
        LOG.info("In Delete fw data is %s", data)
        tenant_name = data.get('tenant_name')
        in_serv_node = self.get_in_srvc_node_ip_addr(tenant_id)
        out_serv_node = self.get_out_srvc_node_ip_addr(tenant_id)
        in_seg, in_vlan = self.get_in_seg_vlan(tenant_id)
        out_seg, out_vlan = self.get_out_seg_vlan(tenant_id)

        kw = dict(params=dict(tenant_name=tenant_name,
                              in_vlan=in_vlan, out_vlan=out_vlan,
                              in_ip=in_serv_node, in_mask='255.255.255.0',
                              out_ip=out_serv_node, out_mask='255.255.255.0',
                              intf_in=self.interface_in,
                              intf_out=self.interface_out))
        status = self.asa5585.cleanup(**kw)
        return status

    def modify_fw(self, tenant_id, data):
        LOG.info("In Modify fw data is %s", data)
        return self.asa5585.apply_policy(data)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.agent import dhcp_agent
from neutron.tests import base

from oslo_config import cfg

from networking_cisco.plugins.cisco.cpnr.dhcp_driver import OPTS
from networking_cisco.plugins.cisco.cpnr import dhcpopts
from networking_cisco.plugins.cisco.cpnr import model
from networking_cisco.tests.unit.cisco.cpnr import fake_networks

dhcp_agent.register_options(cfg.CONF)
cfg.CONF.register_opts(OPTS, 'cisco_pnr')


class TestModel(base.BaseTestCase):

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_network_init(self, mock_client):
        net = model.Network.from_neutron(fake_networks.fake_net1)
        self.assertIsInstance(net, model.Network)
        self.assertFalse(mock_client.called)

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_network_create(self, mock_client):
        net = model.Network.from_neutron(fake_networks.fake_net1)
        net.create()

        # Validate call to CpnrClient.update_vpn
        netid = fake_networks.fake_net1.id
        expected = {'name': netid,
                    'description': netid,
                    'id': model.Vpn.net_to_vpn_id(netid),
                    'vpnId': model.Vpn.net_to_vpn_rfc(netid)}
        mock_client.return_value.update_vpn.assert_called_once_with(
            netid, expected)

        # Validate call to CpnrClient.update_view
        viewid = model.View.net_to_view_id(netid)
        expected = {'name': netid,
                    'viewId': viewid,
                    'priority': viewid}
        mock_client.return_value.update_dns_view.assert_called_once_with(
            netid, expected)

        # Validate call to CpnrClient.update_ccm_zone
        expected = {'origin': 'openstacklocal.',
                    'nameservers': {'stringItem': ['localhost.']},
                    'ns': 'localhost.',
                    'person': 'test.example.com.',
                    'serial': '1',
                    'viewId': viewid}
        mock_client.return_value.update_ccm_zone.assert_called_once_with(
            expected['origin'], expected, viewid=viewid)

        # Validate call to CpnrClient.update_ccm_reverse_zone (reuse fw zone)
        expected['origin'] = '9.9.172.in-addr.arpa.'
        expected['description'] = fake_networks.fake_subnet1.id
        mock_client.return_value.\
            update_ccm_reverse_zone.assert_called_once_with(
                expected['origin'], expected, viewid=viewid)

        # Validate call to CpnrClient.update_scope
        range_list = {'RangeItem': [{'start': '172.9.9.9',
                                     'end': '172.9.9.9'}]}
        policy = model.Policy.from_neutron_subnet(
            fake_networks.fake_net1, fake_networks.fake_subnet1)
        expected = {'name': fake_networks.fake_subnet1.id,
                    'vpnId': model.Vpn.net_to_vpn_id(netid),
                    'subnet': '172.9.9.0/24',
                    'rangeList': range_list,
                    'restrictToReservations': 'enabled',
                    'embeddedPolicy': policy.data}
        mock_client.return_value.update_scope.assert_called_once_with(
            expected['name'], expected)

        # Validate call to CpnrClient.update_client_entry
        policy = model.Policy.from_neutron_port(
            fake_networks.fake_net1, fake_networks.fake_port1)
        netportid = "%s+%s" % (netid, fake_networks.fake_port1.id)
        expected = {'clientClassName': 'openstack-client-class',
                    'name': '01:ab:12:34:56:78:90:ab:aa:bb:cc:dd:ee:ff',
                    'hostName': 'host-172-9-9-9',
                    'domainName': 'openstacklocal',
                    'reservedAddresses': {'stringItem': ['172.9.9.9']},
                    'embeddedPolicy': policy.data,
                    'userDefined': netportid}
        mock_client.return_value.update_client_entry.assert_called_once_with(
            expected['name'], expected)

        # Validate call to CpnrClient.update_ccm_host
        expected = {'name': 'host-172-9-9-9',
                    'zoneOrigin': 'openstacklocal.',
                    'addrs': {'stringItem': ['172.9.9.9']}}
        mock_client.return_value.update_ccm_host.assert_called_once_with(
            expected['name'], expected,
            viewid=viewid, zoneid=expected['zoneOrigin'])

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_port_add(self, mock_client):
        old = model.Network.from_neutron(fake_networks.fake_net1)
        new = model.Network.from_neutron(fake_networks.fake_net2)
        old.update(new)

        # Validate that only port-related objects updated
        self.assertFalse(mock_client.return_value.update_vpn.called)
        self.assertFalse(mock_client.return_value.update_dns_view.called)
        self.assertFalse(mock_client.return_value.update_ccm_zone.called)
        self.assertFalse(mock_client.return_value.
            update_ccm_reverse_zone.called)

        # Validate call to CpnrClient.update_scope
        range_list = {'RangeItem': [{'start': '172.9.9.9',
                                     'end': '172.9.9.10'}]}
        policy = model.Policy.from_neutron_subnet(
            fake_networks.fake_net2, fake_networks.fake_subnet1)
        expected = {'name': fake_networks.fake_subnet1.id,
                    'vpnId': model.Vpn.net_to_vpn_id(
                        fake_networks.fake_net2.id),
                    'subnet': '172.9.9.0/24',
                    'rangeList': range_list,
                    'restrictToReservations': 'enabled',
                    'embeddedPolicy': policy.data}
        mock_client.return_value.update_scope.assert_called_once_with(
            expected['name'], expected)

        # Validate call to CpnrClient.update_client_entry
        netid = fake_networks.fake_net2.id
        policy = model.Policy.from_neutron_port(
            fake_networks.fake_net2, fake_networks.fake_port2)
        netportid = "%s+%s" % (netid, fake_networks.fake_port2.id)
        expected = {'clientClassName': 'openstack-client-class',
                    'name': '01:ab:12:34:56:78:90:ab:aa:bb:cc:dd:ee:99',
                    'hostName': 'host-172-9-9-10',
                    'domainName': 'openstacklocal',
                    'reservedAddresses': {'stringItem': ['172.9.9.10']},
                    'embeddedPolicy': policy.data,
                    'userDefined': netportid}
        mock_client.return_value.update_client_entry.assert_called_once_with(
            expected['name'], expected)

        # Validate call to CpnrClient.update_ccm_host
        viewid = model.View.net_to_view_id(netid)
        expected = {'name': 'host-172-9-9-10',
                    'zoneOrigin': 'openstacklocal.',
                    'addrs': {'stringItem': ['172.9.9.10']}}
        mock_client.return_value.update_ccm_host.assert_called_once_with(
            expected['name'], expected,
            viewid=viewid, zoneid=expected['zoneOrigin'])

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_port_remove(self, mock_client):
        old = model.Network.from_neutron(fake_networks.fake_net2)
        new = model.Network.from_neutron(fake_networks.fake_net1)
        old.update(new)

        # Validate that only port-related objects updated
        self.assertFalse(mock_client.return_value.delete_vpn.called)
        self.assertFalse(mock_client.return_value.delete_dns_view.called)
        self.assertFalse(mock_client.return_value.delete_ccm_zone.called)
        self.assertFalse(
            mock_client.return_value.delete_ccm_reverse_zone.called)

        # Validate call to CpnrClient.update_scope
        range_list = {'RangeItem': [{'start': '172.9.9.9',
                                     'end': '172.9.9.9'}]}
        policy = model.Policy.from_neutron_subnet(
            fake_networks.fake_net1, fake_networks.fake_subnet1)
        expected = {'name': fake_networks.fake_subnet1.id,
                    'vpnId': model.Vpn.net_to_vpn_id(
                        fake_networks.fake_net1.id),
                    'subnet': '172.9.9.0/24',
                    'rangeList': range_list,
                    'restrictToReservations': 'enabled',
                    'embeddedPolicy': policy.data}
        mock_client.return_value.update_scope.assert_called_once_with(
            expected['name'], expected)

        # Validate call to CpnrClient.delete_client_entry
        mock_client.return_value.delete_client_entry.assert_called_once_with(
            '01:ab:12:34:56:78:90:ab:aa:bb:cc:dd:ee:99')

        # Validate call to CpnrClient.release_address
        netid = fake_networks.fake_net2.id
        vpnid = model.Vpn.net_to_vpn_id(netid)
        mock_client.return_value.release_address('172.9.9.10', vpnid)

        # Validate call to CpnrClient.delete_ccm_host
        viewid = model.View.net_to_view_id(netid)
        mock_client.return_value.delete_ccm_host.assert_called_once_with(
            'host-172-9-9-10', viewid=viewid, zoneid='openstacklocal.')

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_network_delete(self, mock_client):
        net = model.Network.from_neutron(fake_networks.fake_net1)
        net.delete()

        # Validate call to CpnrClient.delete_vpn
        netid = fake_networks.fake_net1.id
        mock_client.return_value.delete_vpn.assert_called_once_with(netid)

        # Validate call to CpnrClient.delete_view
        viewid = model.View.net_to_view_id(netid)
        mock_client.return_value.delete_dns_view.assert_called_once_with(netid)

        # Validate call to CpnrClient.delete_ccm_zone
        mock_client.return_value.delete_ccm_zone.assert_called_once_with(
            'openstacklocal.', viewid=viewid)

        # Validate call to CpnrClient.delete_ccm_reverse_zone
        mock_client.return_value.\
            delete_ccm_reverse_zone.assert_called_once_with(
                '9.9.172.in-addr.arpa.', viewid=viewid)

        # Validate call to CpnrClient.delete_scope
        mock_client.return_value.delete_scope.assert_called_once_with(
            fake_networks.fake_subnet1.id)

        # Validate call to CpnrClient.delete_client_entry
        mock_client.return_value.delete_client_entry.assert_called_once_with(
            '01:ab:12:34:56:78:90:ab:aa:bb:cc:dd:ee:ff')

        # Validate call to CpnrClient.release_address
        vpnid = model.Vpn.net_to_vpn_id(netid)
        mock_client.return_value.release_address('172.9.9.9', vpnid)

        # Validate call to CpnrClient.delete_ccm_host
        mock_client.return_value.delete_ccm_host.assert_called_once_with(
            'host-172-9-9-9', viewid=viewid, zoneid='openstacklocal.')

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_reload(self, mock_client):
        mock_client.return_value.reload_needed.return_value = False
        self.assertFalse(model.reload_needed())
        mock_client.return_value.reload_needed.return_value = True
        self.assertTrue(model.reload_needed())

        mock_client.return_value.get_version.return_value = "1"
        mock_client.return_value.get_dhcp_server.return_value = {
            'name': 'fake'}
        mock_client.return_value.get_dns_server.return_value = {
            'name': 'fake'}
        model.reload_server()
        self.assertTrue(mock_client.return_value.reload_server.called)

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_get_version(self, mock_client):
        (mock_client.return_value.get_version.
            return_value) = "CPNR Version 8.3.3"
        ver = model.get_version()
        self.assertEqual('8.3.3', ver)

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_recover_networks(self, mock_client):
        # Setup return values for get functions
        net = model.Network.from_neutron(fake_networks.fake_net2)
        mock_client.return_value.get_vpns.return_value = [net.vpn.data]
        mock_client.return_value.get_scopes.return_value = \
            [s.data for s in net.scopes.values()]
        mock_client.return_value.get_client_entries.return_value = \
            [ce.data for ce in net.client_entries.values()]
        mock_client.return_value.get_dns_views.return_value = [net.view.data]
        mock_client.return_value.get_ccm_zones.return_value = \
            [fz.data for fz in net.forward_zones.values()]
        mock_client.return_value.get_ccm_reverse_zones.return_value = \
            [rz.data for rz in net.reverse_zones.values()]
        mock_client.return_value.get_ccm_hosts.return_value = \
            [h.data for h in net.hosts.values()]

        # Extract key identifiers
        netid = fake_networks.fake_net2.id
        vpnid = net.vpn.data['id']
        viewid = net.view.data['viewId']
        zoneid = 'openstacklocal.'

        # Invoke recover_networks function
        networks = model.recover_networks()
        self.assertIn(netid, networks)
        rec = networks[netid]

        # Validate get functions are called as expected
        mock_client.return_value.get_vpns.assert_called_once_with()
        mock_client.return_value.get_scopes.assert_called_once_with(vpnid)
        mock_client.return_value.get_client_entries.assert_called_once_with()
        mock_client.return_value.get_dns_views.assert_called_once_with()
        mock_client.return_value.\
            get_ccm_zones.assert_called_once_with(viewid=viewid)
        mock_client.return_value.get_ccm_reverse_zones.assert_called_once_with(
            viewid=viewid)
        mock_client.return_value.get_ccm_hosts(viewid=viewid, zoneid=zoneid)

        # Validate that recover_networks returned correct data
        self.assertEqual(net.vpn.data, rec.vpn.data)
        self.assertEqual(net.view.data, rec.view.data)
        for scopeid in net.scopes:
            self.assertIn(scopeid, rec.scopes)
            self.assertEqual(net.scopes[scopeid].data,
                             rec.scopes[scopeid].data)
        for clientid in net.client_entries:
            self.assertIn(clientid, rec.client_entries)
            self.assertEqual(net.client_entries[clientid].data,
                             rec.client_entries[clientid].data)
        for fzid in net.forward_zones:
            self.assertIn(fzid, rec.forward_zones)
            self.assertEqual(net.forward_zones[fzid].data,
                             rec.forward_zones[fzid].data)
        for rzid in net.reverse_zones:
            self.assertIn(rzid, rec.reverse_zones)
            self.assertEqual(net.reverse_zones[rzid].data,
                             rec.reverse_zones[rzid].data)
        for hostid in net.hosts:
            self.assertIn(hostid, rec.hosts)
            self.assertEqual(net.hosts[hostid].data,
                             rec.hosts[hostid].data)

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_policy_from_port(self, mock_client):
        policy = model.Policy.from_neutron_port(fake_networks.fake_net1,
                                                fake_networks.fake_port1)
        opts_list = fake_networks.fake_port1.extra_dhcp_opts
        opt_list_pnr_format = [dhcpopts.format_for_pnr(opts_list[i].opt_name,
                                                       opts_list[i].opt_value)
                               for i in range(len(opts_list))]
        expected = {'OptionItem': opt_list_pnr_format}
        self.assertEqual(expected, policy.data['optionList'])

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_policy_from_subnet(self, mock_client):
        fake_network = fake_networks.fake_net1
        fake_subnet = fake_networks.fake_subnet1
        policy = model.Policy.from_neutron_subnet(fake_network, fake_subnet)
        # DNS servers and static routes should correspond to values in
        # fake_networks.fake_subnet1
        fake_policy_opts = [('routers', fake_subnet.gateway_ip),
                            ('domain-name-servers', '8.8.8.8'),
                            ('classless-static-routes',
                                '18:28:00:01:28:00:00:02'),
                            ('dhcp-lease-time',
                             str(cfg.CONF.dhcp_lease_duration)),
                            ('domain-name', cfg.CONF.dns_domain)]
        policy_list_pnr_format = [dhcpopts.format_for_pnr(name, val)
                                  for name, val in fake_policy_opts]
        expected = {'OptionItem': policy_list_pnr_format}
        self.assertEqual(expected, policy.data['optionList'])

    @mock.patch.object(model, "_get_client", autospec=True)
    def test_scope_from_subnet(self, mock_client):
        policy = model.Policy.from_neutron_subnet(
            fake_networks.fake_net3, fake_networks.fake_subnet1)
        range_list = {'RangeItem': [{'start': '172.9.9.11',
                                     'end': '172.9.9.13'},
                                    {'start': '172.9.9.18',
                                     'end': '172.9.9.18'}]}
        expected = {'name': fake_networks.fake_subnet1.id,
                    'vpnId': model.Vpn.net_to_vpn_id(
                        fake_networks.fake_net3.id),
                    'subnet': '172.9.9.0/24',
                    'rangeList': range_list,
                    'restrictToReservations': 'enabled',
                    'embeddedPolicy': policy.data}
        scope = model.Scope.from_neutron(fake_networks.fake_net3,
                                         fake_networks.fake_subnet1)
        self.assertEqual(expected, scope.data)



# Copyright (c) 2017-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Implements a Nexus-OS NETCONF over SSHv2 API Client
"""

import re
import time

from oslo_log import log as logging

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    exceptions as cexc)
from networking_cisco.ml2_drivers.nexus import (
    nexus_base_network_driver as basedrvr)
from networking_cisco.ml2_drivers.nexus import (
    nexus_db_v2 as nxos_db)
from networking_cisco.ml2_drivers.nexus import (
    nexus_helpers as nexus_help)
from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_client as client)
from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_snippets as snipp)

LOG = logging.getLogger(__name__)

TRUNK_MODE_NOT_FOUND = ("Found trunk vlans but switchport mode is not "
"trunk on Nexus switch %s interface %s. Recheck Nexus Switch config.")


class CiscoNexusRestapiDriver(basedrvr.CiscoNexusBaseDriver):
    """Nexus Driver Restapi Class."""
    def __init__(self, nexus_switches):
        super(CiscoNexusRestapiDriver, self).__init__(nexus_switches)
        credentials = self._build_credentials(self.nexus_switches)
        self.client = self._import_client(credentials)
        self.nxapi_client = self._get_nxapi_client(credentials)
        LOG.debug("ML2 Nexus RESTAPI Drivers initialized.")

    def _import_client(self, credentials):
        """Import the local RESTAPI client module.

        This method was created to mirror original ssh driver so
        test code was in sync.

        """

        return client.CiscoNexusRestapiClient(credentials)

    def _get_nxapi_client(self, credentials):
        """Get the local NXAPI CLI client module.

        :param credentials: contains switch user/password
        :returns: driver to send Nexus CLI cmds via NXAPI.
        """

        return client.CiscoNexusRestapiClient(
            credentials, request_cookie=False)

    def _build_credentials(self, nexus_switches):
        """Build credential table for Rest API Client.

        :param nexus_switches: switch config
        :returns credentials: switch credentials list
        """
        credentials = {}
        for switch_ip, attrs in nexus_switches.items():
            credentials[switch_ip] = (
                attrs[const.USERNAME], attrs[const.PASSWORD],
                attrs[const.HTTPS_VERIFY], attrs[const.HTTPS_CERT],
                None)
            if not attrs[const.HTTPS_VERIFY]:
                LOG.warning("HTTPS Certificate verification is "
                            "disabled. Your connection to Nexus "
                            "Switch %(ip)s is insecure.",
                            {'ip': switch_ip})
        return credentials

    def _get_user_port_channel_config(self, switch_ip, vpc_nbr):
        #Looks for optional user port channel config
        #
        #:param switch_ip: switch to get vpc_nbr user config
        #:param vpc_nbr: which vpc_nbr to apply user config
        #:returns: user configured nexus port channel commands
        #          or None if not present

        def insert_space(matchobj):
            # Command output format must be cmd1 ;cmd2 ; cmdn
            # and not cmd1;cmd2;cmdn or config will fail in Nexus.
            # This does correct formatting.
            test = matchobj.group(0)
            return test[0] + ' ;'

        ucmds = self.nexus_switches[switch_ip].get(const.IF_PC)
        if ucmds:
            ucmds = re.sub("\w;", insert_space, ucmds)
            prefix = 'int port-channel %d ;' % vpc_nbr
            ucmds = ''.join((prefix, ucmds))
        return ucmds

    def get_interface_switch(self, nexus_host,
                             intf_type, interface):
        """Get the interface data from host.

        :param nexus_host: IP address of Nexus switch
        :param intf_type:  String which specifies interface type.
                           example: ethernet
        :param interface:  String indicating which interface.
                           example: 1/19
        :returns response: Returns interface data
        """

        if intf_type == "ethernet":
            path_interface = "phys-[eth" + interface + "]"
        else:
            path_interface = "aggr-[po" + interface + "]"

        action = snipp.PATH_IF % path_interface

        starttime = time.time()
        response = self.client.rest_get(action, nexus_host)
        self.capture_and_print_timeshot(starttime, "getif",
            switch=nexus_host)
        LOG.debug("GET call returned interface %(if_type)s %(interface)s "
            "config", {'if_type': intf_type, 'interface': interface})
        return response

    def _get_interface_switch_trunk_present(
        self, nexus_host, intf_type, interface):
        """Check if 'switchport trunk' configs present.

        :param nexus_host:   IP address of Nexus switch
        :param intf_type:    String which specifies interface type.
                             example: ethernet
        :param interface:    String indicating which interface.
                             example: 1/19
        :returns mode_found: True if 'trunk mode' present
        :returns vlan_configured: True if trunk allowed vlan list present
        """
        result = self.get_interface_switch(nexus_host, intf_type, interface)

        if_type = 'l1PhysIf' if intf_type == "ethernet" else 'pcAggrIf'
        if_info = result['imdata'][0][if_type]

        mode_found = False
        try:
            mode_cfg = if_info['attributes']['mode']
            if mode_cfg == "trunk":
                mode_found = True
        except Exception:
            pass

        vlan_configured = False
        try:
            vlan_list = if_info['attributes']['trunkVlans']
            if vlan_list != const.UNCONFIGURED_VLAN:
                vlan_configured = True
        except Exception:
            pass

        return mode_found, vlan_configured

    def add_ch_grp_to_interface(
        self, nexus_host, if_type, port, ch_grp):
        """Applies channel-group n to ethernet interface."""

        if if_type != "ethernet":
            LOG.error("Unexpected interface type %(iftype)s when "
                      "adding change group", {'iftype': if_type})
            return

        starttime = time.time()

        path_snip = snipp.PATH_ALL
        path_interface = "phys-[eth" + port + "]"

        body_snip = snipp.BODY_ADD_CH_GRP % (ch_grp, ch_grp, path_interface)

        self.send_edit_string(nexus_host, path_snip, body_snip)

        self.capture_and_print_timeshot(
            starttime, "add_ch_group",
            switch=nexus_host)

    def delete_ch_grp_to_interface(
        self, nexus_host, if_type, port, ch_grp):
        """Removes channel-group n from ethernet interface."""

        if if_type != "ethernet":
            LOG.error("Unexpected interface type %(iftype)s when "
                      "deleting change group", {'iftype': if_type})
            return

        starttime = time.time()

        path_snip = snipp.PATH_ALL

        path_interface = "phys-[eth" + port + "]"
        body_snip = snipp.BODY_DEL_CH_GRP % (ch_grp, path_interface)

        self.send_edit_string(nexus_host, path_snip, body_snip)

        self.capture_and_print_timeshot(
            starttime, "del_ch_group",
            switch=nexus_host)

    def _apply_user_port_channel_config(self, nexus_host, vpc_nbr):
        """Adds STP and no lacp suspend config to port channel. """

        cli_cmds = self._get_user_port_channel_config(nexus_host, vpc_nbr)
        if cli_cmds:
            self._send_cli_conf_string(nexus_host, cli_cmds)
        else:
            vpc_str = str(vpc_nbr)
            path_snip = snipp.PATH_ALL
            body_snip = snipp.BODY_ADD_PORT_CH_P2 % (vpc_str, vpc_str)
            self.send_edit_string(nexus_host, path_snip, body_snip)

    def create_port_channel(self, nexus_host, vpc_nbr):
        """Creates port channel n on Nexus switch."""

        starttime = time.time()

        vpc_str = str(vpc_nbr)
        path_snip = snipp.PATH_ALL
        body_snip = snipp.BODY_ADD_PORT_CH % (vpc_str, vpc_str, vpc_str)

        self.send_edit_string(nexus_host, path_snip, body_snip)

        self._apply_user_port_channel_config(nexus_host, vpc_nbr)

        self.capture_and_print_timeshot(
            starttime, "create_port_channel",
            switch=nexus_host)

    def delete_port_channel(self, nexus_host, vpc_nbr):
        """Deletes delete port channel on Nexus switch."""

        starttime = time.time()

        path_snip = snipp.PATH_ALL
        body_snip = snipp.BODY_DEL_PORT_CH % (vpc_nbr)

        self.send_edit_string(nexus_host, path_snip, body_snip)

        self.capture_and_print_timeshot(
            starttime, "delete_port_channel",
            switch=nexus_host)

    def _get_port_channel_group(self, nexus_host, intf_type, interface):
        """Look for 'channel-group x' config and return x.

        :param nexus_host: IP address of Nexus switch
        :param intf_type:  String which specifies interface type.
                           example: ethernet
        :param interface:  String indicating which interface.
                           example: 1/19
        :returns pc_group: Returns port channel group if
                           present else 0
        """

        ch_grp = 0

        # channel-group only applied to ethernet,
        # otherwise, return 0
        if intf_type != 'ethernet':
            return ch_grp

        match_key = "eth" + interface

        action = snipp.PATH_GET_PC_MEMBERS

        starttime = time.time()
        result = self.client.rest_get(action, nexus_host)
        self.capture_and_print_timeshot(starttime, "getpc",
            switch=nexus_host)
        try:
            for pcmbr in result['imdata']:
                mbr_data = pcmbr['pcRsMbrIfs']['attributes']
                if mbr_data['tSKey'] == match_key:
                    _, nbr = mbr_data['parentSKey'].split("po")
                    ch_grp = int(nbr)
                    break
        except Exception:
            pass

        LOG.debug("GET interface %(key)s port channel is %(pc)d",
            {'key': match_key, 'pc': ch_grp})

        return ch_grp

    def _replace_interface_ch_grp(self, interfaces, i, ch_grp):
        # Substitute content of ch_grp in an entry in interface list.
        interfaces[i] = interfaces[i][:-1] + (ch_grp,)

    def _build_host_list_and_verify_chgrp(self, interfaces):
        # Gathers preliminary information for baremetal init
        # and validates ch_grp in db and nexus.
        #
        # 1) Build host ip list from interfaces,
        # 2) first interface sets learned/alloc attribute for
        #    remaining interfaces,
        # 3) check consistent of previously saved chgrp,
        # 4) and check consistency of content in Nexus config

        # :param interfaces:  one or more list of interfaces
        # :returns learned: whether change group learned for set
        #                   of interfaces
        # :returns nexus_ip_list: list of nexus ip address in interfaces.
        #

        # build host list in case you need to allocate vpc_id
        prev_ch_grp = -1
        nexus_ip_list = set()
        learned = False
        max_ifs = len(interfaces)
        for i, (nexus_host, intf_type, nexus_port,
            is_native, ch_grp) in enumerate(interfaces):

            nexus_ip_list.add(nexus_host)
            # if a simple ethernet case, nothing else to do.
            if max_ifs == 1:
                continue

            learned_ch_grp = self._get_port_channel_group(
                nexus_host, intf_type, nexus_port)
            # if first entry
            if prev_ch_grp == -1:
                first_ch_grp = ch_grp
                if ch_grp == 0:
                    save_1st_if = (nexus_host + ', ' +
                        intf_type + ':' + nexus_port + ', ')
                    if learned_ch_grp > 0:
                        learned = True
                        first_ch_grp = learned_ch_grp
            else:
                # Verify ch_grps in Nexus driver port db are consistent
                if prev_ch_grp != ch_grp:
                    LOG.warning("Inconsistent change group stored in "
                        "Nexus port entry data base. Saw %(ch_grp)d expected "
                        "%(p_ch_grp)d for switch %(switch)s interface "
                        "%(intf)s. Updating db.",
                        {'ch_grp': ch_grp,
                         'p_ch_grp': prev_ch_grp,
                         'switch': nexus_host,
                         'intf': nexus_help.format_interface_name(
                             intf_type, nexus_port)})
                # Verify ch_grps on Nexus switches are consistent
                if learned_ch_grp != first_ch_grp:
                    this_if = (nexus_host + ', ' + intf_type +
                        ':' + nexus_port +
                        ', vpc=' + str(learned_ch_grp))
                    if learned:
                        # Learned ch_grp not consistent between
                        # interfaces in this set.
                        raise cexc.NexusVPCLearnedNotConsistent(
                            first=save_1st_if + 'vpc=' + str(first_ch_grp),
                            second=this_if)
                    else:
                        # Learned ch_grp for this interface while
                        # first is non-learned is not consistent.
                        raise cexc.NexusVPCExpectedNoChgrp(
                            first=save_1st_if + 'vpc=None',
                            second=this_if)
            if learned:
                self._replace_interface_ch_grp(
                    interfaces, i, learned_ch_grp)

            prev_ch_grp = ch_grp

        return learned, list(nexus_ip_list)

    def _config_new_baremetal_portchannel(self, ch_grp, nexus_host,
                                          if_type, nexus_port):
        # Handles config port-channel creation for baremetal event.
        #
        # Creates port-channel then applies the channel-group
        # to ethernet interface.
        # :param ch_grp:        vpcid/ch_grp to use
        # :param nexus_host:    ip of first switch in event
        # :param if_type:       interface type in event
        # :param nexus_port:    interface port in event

        self.create_port_channel(nexus_host, ch_grp)
        self.add_ch_grp_to_interface(
            nexus_host, if_type, nexus_port, ch_grp)

    def _get_new_baremetal_portchannel_id(self, nexus_ip_list):
        # Gets initial channel group id for list of nexus switches.
        #
        # :param nexus_ip_list: list of nexus switch to allocate vpcid
        # :returns ch_grp:      vpc_id allocated

        # When allocating vpcid, it must be unique for
        # all switches in the set; else error out.
        ch_grp = nxos_db.alloc_vpcid(nexus_ip_list)
        if ch_grp == 0:
            nexus_ip_list.sort()
            ip_str_list = ','.join('%s' % ip for ip in nexus_ip_list)
            raise cexc.NexusVPCAllocFailure(
                switches=ip_str_list)

        return ch_grp

    def _configure_learned_port_channel(self, nexus_ip_list, ch_grp):
        # Handle baremetal interfaces when vpc-id was learned from Nexus.
        #
        #:param nexus_ip_list: list of nexus switch to allocate vpcid
        #:param ch_grp: learned ch_grp

        try:
            nxos_db.update_vpc_entry(
                nexus_ip_list, ch_grp, True, True)
        except cexc.NexusVPCAllocNotFound:
            # Valid to get this error if learned ch_grp
            # not part of configured vpc_pool
            pass

    def initialize_baremetal_switch_interfaces(self, interfaces):
        """Initialize Nexus interfaces and for initial baremetal event.

        This get/create port channel number, applies channel-group to
        ethernet interface, and initializes trunking on interface.

        :param interfaces: Receive a list of interfaces containing:
            nexus_host: IP address of Nexus switch
            intf_type: String which specifies interface type. example: ethernet
            interface: String indicating which interface. example: 1/19
            is_native: Whether native vlan must be configured.
            ch_grp: May replace port channel to each entry.  channel number is
            0 if none
        """
        if not interfaces:
            return

        max_ifs = len(interfaces)
        starttime = time.time()

        learned, nexus_ip_list = self._build_host_list_and_verify_chgrp(
            interfaces)
        if not nexus_ip_list:
            return

        if max_ifs > 1:
            # update vpc db with learned vpcid or get new one.
            if learned:
                ch_grp = interfaces[0][-1]
                self._configure_learned_port_channel(
                    nexus_ip_list, ch_grp)
            else:
                ch_grp = self._get_new_baremetal_portchannel_id(nexus_ip_list)
        else:
            ch_grp = 0

        for i, (nexus_host, intf_type, nexus_port, is_native,
            ch_grp_saved) in enumerate(interfaces):

            if max_ifs > 1:
                if learned:
                    ch_grp = ch_grp_saved
                else:
                    self._config_new_baremetal_portchannel(
                        ch_grp, nexus_host, intf_type, nexus_port)
                    self._replace_interface_ch_grp(interfaces, i, ch_grp)

                # init port-channel instead of the provided ethernet
                intf_type = 'port-channel'
                nexus_port = str(ch_grp)
            else:
                self._replace_interface_ch_grp(interfaces, i, ch_grp)

            trunk_mode_present, vlan_present = (
                self._get_interface_switch_trunk_present(
                    nexus_host, intf_type, nexus_port))
            if not vlan_present:
                self.send_enable_vlan_on_trunk_int(
                    nexus_host, "", intf_type, nexus_port, False,
                    not trunk_mode_present)
            elif not trunk_mode_present:
                LOG.warning(TRUNK_MODE_NOT_FOUND, nexus_host,
                            nexus_help.format_interface_name(
                                intf_type, nexus_port))

        self.capture_and_print_timeshot(
            starttime, "init_bmif",
            switch=nexus_host)

    def initialize_all_switch_interfaces(self, interfaces,
                                         switch_ip=None, replay=True):
        """Configure Nexus interface and get port channel number.

        Called during switch replay or just init if no replay
        is configured.  For latter case, only configured interfaces
        are affected by this method.

        During switch replay, the change group from the
        host mapping data base is used.  There is no attempt
        to relearn port-channel from the Nexus switch.  What
        we last knew it to be will persist.

        :param interfaces:  List of interfaces for a given switch.
                            ch_grp can be altered as last arg
                            to each interface. If no ch_grp,
                            this arg will be zero.
        :param switch_ip: IP address of Nexus switch
        :param replay: Whether in replay path
        """
        if not interfaces:
            return

        starttime = time.time()

        if replay:
            try:
                vpcs = nxos_db.get_active_switch_vpc_allocs(switch_ip)
            except cexc.NexusVPCAllocNotFound:
                vpcs = []
            for vpc in vpcs:
                # if this is an allocated vpc, then recreate it
                if not vpc.learned:
                    self.create_port_channel(switch_ip, vpc.vpc_id)

        for i, (nexus_host, intf_type, nexus_port, is_native,
            ch_grp) in enumerate(interfaces):
            if replay and ch_grp != 0:
                try:
                    vpc = nxos_db.get_switch_vpc_alloc(switch_ip, ch_grp)
                    self.add_ch_grp_to_interface(
                        nexus_host, intf_type, nexus_port, ch_grp)
                except cexc.NexusVPCAllocNotFound:
                    pass
                # if channel-group exists, switch to port-channel
                # instead of the provided ethernet interface
                intf_type = 'port-channel'
                nexus_port = str(ch_grp)

            #substitute content of ch_grp
            no_chgrp_len = len(interfaces[i]) - 1
            interfaces[i] = interfaces[i][:no_chgrp_len] + (ch_grp,)

            trunk_mode_present, vlan_present = (
                self._get_interface_switch_trunk_present(
                    nexus_host, intf_type, nexus_port))
            if not vlan_present:
                self.send_enable_vlan_on_trunk_int(
                    nexus_host, "", intf_type, nexus_port, False,
                    not trunk_mode_present)
            elif not trunk_mode_present:
                LOG.warning(TRUNK_MODE_NOT_FOUND, nexus_host,
                            nexus_help.format_interface_name(
                                intf_type, nexus_port))

        self.capture_and_print_timeshot(
            starttime, "get_allif",
            switch=nexus_host)

    def get_nexus_type(self, nexus_host):
        """Given the nexus host, get the type of Nexus switch.

        :param nexus_host: IP address of Nexus switch
        :returns: Nexus type
        """

        starttime = time.time()
        response = self.client.rest_get(
            snipp.PATH_GET_NEXUS_TYPE, nexus_host)
        self.capture_and_print_timeshot(
            starttime, "gettype",
            switch=nexus_host)

        result = ''
        if response:
            try:
                result = response['imdata'][0]["eqptCh"]['attributes']['descr']
            except Exception:
                pass
            nexus_type = re.findall(
                "Nexus\s*(\d)\d+\s*[0-9A-Z]+\s*"
                "[cC]hassis",
                result)
            if len(nexus_type) > 0:
                LOG.debug("GET call returned Nexus type %d",
                    int(nexus_type[0]))
                return int(nexus_type[0])

        LOG.debug("GET call failed to return Nexus type. Received %s.",
                  result)
        return -1

    def start_create_vlan(self):
        """Returns REST API path and config start."""

        return snipp.PATH_ALL, snipp.BODY_VLAN_ALL_BEG

    def end_create_vlan(self, conf_str):
        """Returns current config + end of config."""

        conf_str += snipp.BODY_VLAN_ALL_END
        return conf_str

    def get_create_vlan(self, nexus_host, vlanid, vni, conf_str):
        """Returns an XML snippet for create VLAN on a Nexus Switch."""

        starttime = time.time()

        if vni:
            body_snip = snipp.BODY_VXLAN_ALL_INCR % (vlanid, vni)
        else:
            body_snip = snipp.BODY_VLAN_ALL_INCR % vlanid
        conf_str += body_snip + snipp.BODY_VLAN_ALL_CONT

        self.capture_and_print_timeshot(
            starttime, "get_create_vlan",
            switch=nexus_host)

        return conf_str

    def set_all_vlan_states(self, nexus_host, vlanid_range):
        """Set the VLAN states to active."""

        starttime = time.time()

        if not vlanid_range:
            LOG.warning("Exiting set_all_vlan_states: "
                        "No vlans to configure")
            return

        # Eliminate possible whitespace and separate vlans by commas
        vlan_id_list = re.sub(r'\s', '', vlanid_range).split(',')
        if not vlan_id_list or not vlan_id_list[0]:
            LOG.warning("Exiting set_all_vlan_states: "
                        "No vlans to configure")
            return

        path_str, body_vlan_all = self.start_create_vlan()
        while vlan_id_list:
            rangev = vlan_id_list.pop(0)
            if '-' in rangev:
                fr, to = rangev.split('-')
                max = int(to) + 1
                for vlan_id in range(int(fr), max):
                    body_vlan_all = self.get_create_vlan(
                        nexus_host, vlan_id, 0, body_vlan_all)
            else:
                body_vlan_all = self.get_create_vlan(
                    nexus_host, rangev, 0, body_vlan_all)

        body_vlan_all = self.end_create_vlan(body_vlan_all)
        self.send_edit_string(
            nexus_host, path_str, body_vlan_all)

        self.capture_and_print_timeshot(
            starttime, "set_all_vlan_states",
            switch=nexus_host)

    def create_vlan(self, nexus_host, vlanid, vni):
        """Given switch, vlanid, vni, Create a VLAN on Switch."""

        starttime = time.time()
        path_snip, body_snip = self.start_create_vlan()
        body_snip = self.get_create_vlan(nexus_host, vlanid, vni, body_snip)
        body_snip = self.end_create_vlan(body_snip)

        self.send_edit_string(nexus_host, path_snip, body_snip)

        self.capture_and_print_timeshot(
            starttime, "create_vlan_seg",
            switch=nexus_host)

    def delete_vlan(self, nexus_host, vlanid):
        """Delete a VLAN on Nexus Switch given the VLAN ID."""
        starttime = time.time()

        path_snip = snipp.PATH_VLAN % vlanid
        self.client.rest_delete(path_snip, nexus_host)

        self.capture_and_print_timeshot(
            starttime, "del_vlan",
            switch=nexus_host)

    def _get_vlan_body_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                    interface, is_native, is_delete,
                                    add_mode):
        """Prepares an XML snippet for VLAN on a trunk interface.

        :param nexus_host: IP address of Nexus switch
        :param vlanid:     Vlanid(s) to add to interface
        :param intf_type:  String which specifies interface type.
                           example: ethernet
        :param interface:  String indicating which interface.
                           example: 1/19
        :param is_native:  Is native vlan config desired?
        :param is_delete:  Is this a delete operation?
        :param add_mode:   Add mode trunk
        :returns           path_snippet, body_snippet
        """

        starttime = time.time()

        LOG.debug("NexusDriver get if body config for host %s: "
                  "if_type %s port %s",
                  nexus_host, intf_type, interface)
        if intf_type == "ethernet":
            body_if_type = "l1PhysIf"
            path_interface = "phys-[eth" + interface + "]"
        else:
            body_if_type = "pcAggrIf"
            path_interface = "aggr-[po" + interface + "]"

        path_snip = (snipp.PATH_IF % (path_interface))

        mode = snipp.BODY_PORT_CH_MODE if add_mode else ''
        if is_delete:
            increment_it = "-"
            debug_desc = "delif"
            native_vlan = ""
        else:
            native_vlan = 'vlan-' + str(vlanid)
            debug_desc = "createif"
            if vlanid is "":
                increment_it = ""
            else:
                increment_it = "+"

        if is_native:
            body_snip = (snipp.BODY_NATIVE_TRUNKVLAN %
                (body_if_type, mode, increment_it + str(vlanid),
                str(native_vlan)))
        else:
            body_snip = (snipp.BODY_TRUNKVLAN %
                (body_if_type, mode, increment_it + str(vlanid)))

        self.capture_and_print_timeshot(
            starttime, debug_desc,
            switch=nexus_host)

        return path_snip, body_snip

    def disable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                  interface, is_native):
        """Disable a VLAN on a trunk interface."""

        starttime = time.time()

        path_snip, body_snip = self._get_vlan_body_on_trunk_int(
            nexus_host, vlanid, intf_type, interface,
            is_native, True, False)
        self.send_edit_string(nexus_host, path_snip, body_snip)
        self.capture_and_print_timeshot(
            starttime, "delif",
            switch=nexus_host)

    def send_edit_string(self, nexus_host, path_snip, body_snip,
                         check_to_close_session=True):
        """Sends rest Post request to Nexus switch."""

        starttime = time.time()
        LOG.debug("NexusDriver edit config for host %s: path: %s body: %s",
                  nexus_host, path_snip, body_snip)
        self.client.rest_post(path_snip, nexus_host, body_snip)
        self.capture_and_print_timeshot(
            starttime, "send_edit",
            switch=nexus_host)

    def _send_cli_conf_string(self, nexus_host, cli_str):
        """Sends CLI Config commands to Nexus switch using NXAPI."""

        starttime = time.time()
        path_snip = snipp.PATH_USER_CMDS
        body_snip = snipp.BODY_USER_CONF_CMDS % ('1', cli_str)
        LOG.debug("NexusDriver CLI config for host %s: path: %s body: %s",
                  nexus_host, path_snip, body_snip)
        self.nxapi_client.rest_post(path_snip, nexus_host, body_snip)
        self.capture_and_print_timeshot(
            starttime, "send_cliconf",
            switch=nexus_host)

    def send_enable_vlan_on_trunk_int(self, nexus_host, vlanid, intf_type,
                                      interface, is_native, add_mode=False):
        """Gathers and sends an interface trunk XML snippet."""

        path_snip, body_snip = self._get_vlan_body_on_trunk_int(
            nexus_host, vlanid, intf_type, interface,
            is_native, False, add_mode)
        self.send_edit_string(nexus_host, path_snip, body_snip)

    def create_and_trunk_vlan(self, nexus_host, vlan_id, intf_type,
                              nexus_port, vni, is_native):
        """Create VLAN and trunk it on the specified ports."""

        starttime = time.time()

        self.create_vlan(nexus_host, vlan_id, vni)
        LOG.debug("NexusDriver created VLAN: %s", vlan_id)

        if nexus_port:
            self.send_enable_vlan_on_trunk_int(
                nexus_host, vlan_id,
                intf_type, nexus_port,
                is_native)

        self.capture_and_print_timeshot(
            starttime, "create_all",
            switch=nexus_host)

    def enable_vxlan_feature(self, nexus_host, nve_int_num, src_intf):
        """Enable VXLAN on the switch."""

        # Configure the "feature" commands and NVE interface
        # (without "member" subcommand configuration).
        # The Nexus 9K will not allow the "interface nve" configuration
        # until the "feature nv overlay" command is issued and installed.
        # To get around the N9K failing on the "interface nve" command
        # send the two XML snippets down separately.

        starttime = time.time()

        ## Do CLI 'feature nv overlay'
        self.send_edit_string(nexus_host, snipp.PATH_VXLAN_STATE,
                              (snipp.BODY_VXLAN_STATE % "enabled"))

        # Do CLI 'feature vn-segment-vlan-based'
        self.send_edit_string(nexus_host, snipp.PATH_VNSEG_STATE,
                              (snipp.BODY_VNSEG_STATE % "enabled"))

        # Do CLI 'int nve1' to Create nve1
        self.send_edit_string(
            nexus_host,
            (snipp.PATH_NVE_CREATE % nve_int_num),
            (snipp.BODY_NVE_CREATE % nve_int_num))

        # Do CLI 'no shut
        #         source-interface loopback %s'
        # beneath int nve1
        self.send_edit_string(
            nexus_host,
            (snipp.PATH_NVE_CREATE % nve_int_num),
            (snipp.BODY_NVE_ADD_LOOPBACK % ("enabled", src_intf)))
        self.capture_and_print_timeshot(
            starttime, "enable_vxlan",
            switch=nexus_host)

    def disable_vxlan_feature(self, nexus_host):
        """Disable VXLAN on the switch."""

        # Removing the "feature nv overlay" configuration also
        # removes the "interface nve" configuration.

        starttime = time.time()

        # Do CLI 'no feature nv overlay'
        self.send_edit_string(nexus_host, snipp.PATH_VXLAN_STATE,
                              (snipp.BODY_VXLAN_STATE % "disabled"))

        # Do CLI 'no feature vn-segment-vlan-based'
        self.send_edit_string(nexus_host, snipp.PATH_VNSEG_STATE,
                              (snipp.BODY_VNSEG_STATE % "disabled"))

        self.capture_and_print_timeshot(
            starttime, "disable_vxlan",
            switch=nexus_host)

    def create_nve_member(self, nexus_host, nve_int_num, vni, mcast_group):
        """Add a member configuration to the NVE interface."""

        # Do CLI [no] member vni %s mcast-group %s
        # beneath int nve1

        starttime = time.time()

        path = snipp.PATH_VNI_UPDATE % (nve_int_num, vni)
        body = snipp.BODY_VNI_UPDATE % (vni, vni, vni, mcast_group)
        self.send_edit_string(nexus_host, path, body)

        self.capture_and_print_timeshot(
            starttime, "create_nve",
            switch=nexus_host)

    def delete_nve_member(self, nexus_host, nve_int_num, vni):
        """Delete a member configuration on the NVE interface."""

        starttime = time.time()

        path_snip = snipp.PATH_VNI_UPDATE % (nve_int_num, vni)
        self.client.rest_delete(path_snip, nexus_host)

        self.capture_and_print_timeshot(
            starttime, "delete_nve",
            switch=nexus_host)



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from binascii import hexlify
import mock
import socket
import unittest

from networking_cisco.plugins.cisco.cpnr.cpnr_client import UnexpectedError
from networking_cisco.plugins.cisco.cpnr.cpnr_dns_relay_agent import (
    DnsRelayAgent)
from networking_cisco.plugins.cisco.cpnr.cpnr_dns_relay_agent import cfg
from networking_cisco.plugins.cisco.cpnr.cpnr_dns_relay_agent import DnsPacket
from networking_cisco.plugins.cisco.cpnr.cpnr_dns_relay_agent import OPTS


class TestDnsRelayAgent(unittest.TestCase):

    @mock.patch('networking_cisco.plugins.cisco.'
                'cpnr.cpnr_dns_relay_agent.netns')
    @mock.patch('socket.socket')
    def test_open_dns_ext_socket(self,
                                 mock_socket,
                                 mock_netns):
        cfg.CONF.register_opts(OPTS, 'cisco_pnr')
        relay = DnsRelayAgent()

        mock_netns.iflist.return_value = []
        mock_netns.iflist.return_value.append(('lo', '127.0.0.1', '255.0.0.0'))

        sock = mock_socket.return_value
        sock.getsockname.return_value = ('127.0.0.1', 123456)

        sock, addr, port = relay._open_dns_ext_socket()

        mock_socket.assert_has_calls([
            mock.call(socket.AF_INET, socket.SOCK_DGRAM),
            mock.call().bind(('127.0.0.1', 0)),
            mock.call().getsockname(),
            mock.call().connect(('127.0.0.1', 53))]
        )

        # check exception thrown if no interfaces
        with self.assertRaises(UnexpectedError):
            mock_netns.iflist.return_value = []
            sock, addr, port = relay._open_dns_ext_socket()

        # check exception thrown if no matching interfaces
        with self.assertRaises(UnexpectedError):
            mock_netns.iflist.return_value = []
            mock_netns.iflist.return_value.append(('eth0', '10.0.0.10',
                                                   '255.255.255.0'))
            sock, addr, port = relay._open_dns_ext_socket()

        # check matching interface found if not first in list
        mock_netns.iflist.return_value = []
        mock_netns.iflist.return_value.append(('eth0', '10.0.0.10',
                                               '255.255.255.0'))
        mock_netns.iflist.return_value.append(('lo', '127.0.0.1', '255.0.0.0'))
        sock, addr, port = relay._open_dns_ext_socket()

    @mock.patch('networking_cisco.plugins.cisco.'
                'cpnr.cpnr_dns_relay_agent.netns')
    @mock.patch('socket.socket')
    def test_open_dns_int_socket(self,
                                 mock_socket,
                                 mock_netns):
        cfg.CONF.register_opts(OPTS, 'cisco_pnr')
        relay = DnsRelayAgent()

        mock_netns.iflist.return_value = []
        mock_netns.iflist.return_value.append(('eth0', '10.21.1.13',
                                               '255.255.255.0'))
        sock, addr, port = relay._open_dns_int_socket()

        self.assertTrue(mock_netns.iflist.called, "Failed to call iflist.")

        mock_socket.assert_has_calls([
            mock.call(socket.AF_INET, socket.SOCK_DGRAM),
            mock.call().setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call().bind(('10.21.1.13', 53))]
        )

        # check exception thrown if no interfaces
        with self.assertRaises(UnexpectedError):
            mock_netns.iflist.return_value = []
            sock, addr, port = relay._open_dns_int_socket()

    def test_convert_namespace_to_viewid(self):
        cfg.CONF.register_opts(OPTS, 'cisco_pnr')
        relay = DnsRelayAgent()

        namespace = 'qdhcp-d7c31f74-5d9e-47b7-86f2-64879023c04d'
        viewid = relay._convert_namespace_to_viewid(namespace)
        tmp = 0x64879023c04d & 0x7fffffff
        self.assertEqual(viewid, str(tmp))


class TestDnsPacket(unittest.TestCase):

    def test_parse(self):
        # test regular DNS request
        line = ('84 a5 01 00 00 01 00 00 00 00 00 00 06 72 '
               '65 64 68 61 74 03 63 6f 6d 00 00 01 00 01')
        buf = bytearray.fromhex(line)
        pkt = DnsPacket.parse(buf, 28)
        self.assertEqual(0x84a5, pkt.get_msgid())
        self.assertTrue(pkt.isreq)
        self.assertEqual(0, pkt.arcnt)
        self.assertEqual(0, pkt.optlen)
        self.assertEqual(28, pkt.txt_insert_pos)

        # test DNS request with EDNS0
        line = ('81 71 01 20 00 01 00 00 00 00 00 01 06 72 65 '
                '64 68 61 74 03 63 6f 6d 00 00 01 00 01 00 00 '
                '29 10 00 00 00 00 00 00 00')
        buf = bytearray.fromhex(line)
        pkt = DnsPacket.parse(buf, 38)
        self.assertEqual(0x8171, pkt.get_msgid())
        self.assertTrue(pkt.isreq)
        self.assertEqual(1, pkt.arcnt)
        self.assertEqual(10, pkt.optlen)
        self.assertEqual(28, pkt.txt_insert_pos)

        # test regular DNS response
        line = ('b6 5e 81 80 00 01 00 01 00 00 00 00 06 72 65 '
                '64 68 61 74 03 63 6f 6d 00 00 01 00 01 c0 0c '
                '00 01 00 01 00 00 00 08 00 04 d1 84 b7 69')
        buf = bytearray.fromhex(line)
        pkt = DnsPacket.parse(buf, 44)
        self.assertEqual(0xb65e, pkt.get_msgid())
        self.assertFalse(pkt.isreq)
        self.assertEqual(0, pkt.arcnt)
        self.assertEqual(0, pkt.optlen)
        self.assertEqual(-1, pkt.txt_insert_pos)

    def test_set_viewid(self):
        pkt = DnsPacket()
        pkt.set_viewid('123456789')
        self.assertEqual(pkt.viewid, '123456789')

    def test_data(self):
        # call with regular DNS request
        line = ('84 a5 01 00 00 01 00 00 00 00 00 00 06 72 '
               '65 64 68 61 74 03 63 6f 6d 00 00 01 00 01')
        buf = bytearray.fromhex(line)
        pktbuf = bytearray(4096)
        pktbuf[0:len(buf)] = buf
        pkt = DnsPacket.parse(pktbuf, 28)
        pkt.set_viewid('123456')
        mod_buf = pkt.data()
        self.assertEqual(pkt.arcnt, 1)
        hextxtstr = hexlify(DnsPacket.TXT_RR)
        hexstr = hexlify(mod_buf)
        self.assertNotEqual(-1, hexstr.find(hextxtstr))

        # call with DNS request with EDNS0
        line = ('81 71 01 20 00 01 00 00 00 00 00 01 06 72 65 '
                '64 68 61 74 03 63 6f 6d 00 00 01 00 01 00 00 '
                '29 10 00 00 00 00 00 00 00')
        buf = bytearray.fromhex(line)
        pktbuf = bytearray(4096)
        pktbuf[0:len(buf)] = buf
        pkt = DnsPacket.parse(pktbuf, 38)
        pkt.set_viewid('123456')
        mod_buf = pkt.data()
        self.assertEqual(2, pkt.arcnt)
        hexstr = hexlify(mod_buf)
        self.assertNotEqual(-1, hexstr.find(hextxtstr))

    def test_skip_over_domain_name(self):
        # test skip over name at beginning, end up on ^
        # 4test5cisco3com0^
        bytes = bytearray(b'\x04\x74\x65\x73\x74\x05\x63\x69\x73\x63'
                          b'\x6f\x03\x63\x6f\x6d\x00\x5e')
        pos = DnsPacket.skip_over_domain_name(bytes, 0)
        self.assertEqual(16, pos)
        self.assertEqual('^', chr(bytes[pos]))

        # test skip over name in the middle, end up on ^
        # 2552552552554test5cisco3com0^
        bytes = bytearray(b'\xff\xff\xff\xff\x04\x74\x65\x73\x74\x05\x63'
                          b'\x69\x73\x63\x6f\x03\x63\x6f\x6d\x00\x5e')
        pos = DnsPacket.skip_over_domain_name(bytes, 4)
        self.assertEqual(20, pos)
        self.assertEqual('^', chr(bytes[pos]))

        # test skip over length and pointer at beginning, end up on ^
        bytes = bytearray(b'\xc0\x55\x5e')
        pos = DnsPacket.skip_over_domain_name(bytes, 0)
        self.assertEqual(2, pos)
        self.assertEqual('^', chr(bytes[pos]))

        # test skip over length and pointer in the middle, end up on ^
        bytes = bytearray(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc0\x55\x5e')
        pos = DnsPacket.skip_over_domain_name(bytes, 9)
        self.assertEqual(11, pos)
        self.assertEqual('^', chr(bytes[pos]))



# Copyright 2017 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add hostid as primary key to mapping table

Revision ID: 6454463cea8b
Revises: 73c84db9f299
Create Date: 2017-12-01 18:48:03.531425

"""

# revision identifiers, used by Alembic.
revision = '6454463cea8b'
down_revision = '73c84db9f299'

from alembic import op

from sqlalchemy.engine.reflection import Inspector as insp

ML2_NEXUS_MAPPING_TABLE = 'cisco_ml2_nexus_host_interface_mapping'


def upgrade():

    bind = op.get_bind()

    inspector = insp.from_engine(bind)
    pk_constraint = inspector.get_pk_constraint(ML2_NEXUS_MAPPING_TABLE)
    op.drop_constraint(pk_constraint.get('name'), ML2_NEXUS_MAPPING_TABLE,
                       type_='primary')
    op.create_primary_key(op.f('pk_cisco_ml2_nexus_host_interface_mapping'),
                          ML2_NEXUS_MAPPING_TABLE,
                          ['host_id', 'switch_ip', 'if_id'])



# Copyright 2017 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_utils import uuidutils

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole


_uuid = uuidutils.generate_uuid
HA_INFO = 'ha_info'
ROUTER_ROLE_HA_REDUNDANCY = c_constants.ROUTER_ROLE_HA_REDUNDANCY
ROUTER_ROLE_GLOBAL = c_constants.ROUTER_ROLE_GLOBAL


class CfgAgentTestSupportMixin(object):

    def create_router_port(self, network_uuid, vlan_tag, k, num_subnets,
                           router_id, admin_state_up=True,
                           mac_address='ca:fe:de:ad:be:ef',
                           device_owner=constants.DEVICE_OWNER_ROUTER_INTF,
                           ha_enabled=True, ha_group=1060,
                           is_user_visible=True):
        int_fixed_ips = []
        int_subnets = []
        ha_fixed_ips = []
        sn_ids = sorted([_uuid() for i in range(num_subnets)])
        phy_infc = 'GigabitEthernet0/0/0'
        for i in range(num_subnets):
            sn_id = sn_ids[i]
            if device_owner == constants.DEVICE_OWNER_ROUTER_INTF:
                prefix_len = 24
                host_id = 4
            else:
                prefix_len = 28 if i == 0 else 27
                host_id = 1
            int_fixed_ips.append({'ip_address': '%s.4.%s.%s' % (k, i, host_id),
                                  'prefixlen': prefix_len,
                                  'subnet_id': sn_id})
            int_subnets.append({'id': sn_id,
                                'cidr': '%s.4.%s.0/%s' % (k, i, prefix_len),
                                'gateway_ip': '%s.4.%s.1' % (k, i)})
            ha_fixed_ips.append({'ip_address': '%s.4.%s.254' % (k, i),
                                 'prefixlen': prefix_len,
                                 'subnet_id': sn_id})
        port = {
            'id': _uuid(),
            'network_id': network_uuid,
            'admin_state_up': admin_state_up,
            'device_owner': device_owner,
            'device_id': router_id,
            'fixed_ips': int_fixed_ips,
            'mac_address': mac_address,
            'subnets': int_subnets,
            'hosting_info': {
                'physical_interface': phy_infc,
                'segmentation_id': vlan_tag
            }
        }
        if ha_enabled is True:
            port[HA_INFO] = {'group': ha_group}
            if (device_owner == constants.DEVICE_OWNER_ROUTER_INTF or
                    is_user_visible is False):
                port[HA_INFO]['ha_port'] = {
                    'id': _uuid(),
                    'fixed_ips': ha_fixed_ips,
                    'subnets': int_subnets,
                    'network_id': network_uuid,
                    'admin_state_up': admin_state_up,
                    'device_owner': device_owner,
                    'mac_address': mac_address
                }
            else:
                port[HA_INFO]['ha_port'] = port
        return port

    def prepare_router_data(self, set_gateway=True, enable_snat=None,
                            num_ext_subnets=1, num_internal_ports=1,
                            same_internal_nw=False, is_global=False,
                            ha_enabled=True, is_user_visible=True):
        router_id = _uuid()
        if set_gateway is True:
            ex_gw_port = self.create_router_port(
                _uuid(), 100, 19, num_ext_subnets, router_id,
                device_owner=constants.DEVICE_OWNER_ROUTER_GW)
        else:
            ex_gw_port = None
        int_ports = []
        num_internal_subnets = 1 if is_global is False else num_ext_subnets
        nw_uuid = _uuid()
        vlan_tag = 200
        for j in range(num_internal_ports):
            k = 35 + j
            if ha_enabled is True:
                group_id = 1060 + j
                p = self.create_router_port(
                    nw_uuid, vlan_tag, k, num_internal_subnets, router_id,
                    ha_group=group_id)
            else:
                p = self.create_router_port(
                    nw_uuid, vlan_tag, k, num_internal_subnets, router_id,
                    ha_enabled=False)
            int_ports.append(p)
            if same_internal_nw is False:
                nw_uuid = _uuid()
                vlan_tag += 1

        hosting_device = {
            'id': _uuid(),
            'name': "VM_template",
            'booting_time': 300,
            'host_category': "VM",
            'management_ip_address': '20.0.0.5',
            'protocol_port': 22,
            'credentials': {
                'username': "user",
                'password': "4getme"
            }
        }
        router = {
            'id': router_id,
            'status': 'ACTIVE',
            'admin_state_up': True,
            bc.constants.INTERFACE_KEY: int_ports,
            'routes': [],
            'hosting_device': hosting_device,
            'router_type': ''
        }
        if is_global is True:
            router['gw_port'] = None
            router[routerrole.ROUTER_ROLE_ATTR] = ROUTER_ROLE_GLOBAL
        else:
            router['gw_port'] = ex_gw_port
            if is_user_visible is True:
                router[routerrole.ROUTER_ROLE_ATTR] = None
            else:
                router[routerrole.ROUTER_ROLE_ATTR] = ROUTER_ROLE_HA_REDUNDANCY
        if enable_snat is not None:
            router['enable_snat'] = enable_snat
        if ha_enabled is True:
            ha_details = {
                'priority': 10,
                'redundancy_level': 1,
                'state': 'ACTIVE',
                'type': 'HSRP',
                'redundancy_routers': [{'id': _uuid(),
                                        'priority': 20,
                                        'state': 'STANDBY'}]
            }
            if is_global is True:
                ha_details['redundancy_level'] = 2
                ha_details['redundancy_routers'].append({'id': router_id,
                                                         'priority': 10,
                                                         'state': 'ACTIVE'})
            router[ha.ENABLED] = True
            router[ha.DETAILS] = ha_details
        else:
            router[ha.ENABLED] = False
        return router, int_ports

    def prepare_hosting_device_params(self):
        device_params = {
            'management_ip_address': 'fake_ip',
            'protocol_port': 22,
            'credentials': {
                'user_name': "stack",
                'password': "cisco"
            },
            'timeout': None,
            'id': '0000-1',
            'device_id': 'ASR-1'
        }
        return device_params



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This driver api is adopted from a blueprint implementation what was
# started by Gary Duan, vArmour. That implementation was eventually abandoned.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class L3RouterBaseDriver(object):

    @abc.abstractmethod
    def create_router_precommit(self, context, router_context):
        """Perform operations specific to the router type in preparation for
        the creation of a new router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the new
        router.

        Create a new router, allocating resources as necessary in the
        database. Called inside transaction context on session. Call cannot
        block. Raising an exception will result in a rollback of the current
        transaction.
        """
        pass

    @abc.abstractmethod
    def create_router_postcommit(self, context, router_context):
        """Create a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the new
        router.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause the deletion of
        the resource.
        """
        pass

    @abc.abstractmethod
    def update_router_precommit(self, context, router_context):
        """Perform operations specific to the router type in preparation for
        the update of a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the new
        state of the router, as well as the original state prior to the
        update_router call.

        Update values of a router, updating the associated resources in the
        database. Called inside transaction context on session. Raising an
        exception will result in rollback of the transaction.

        update_router_precommit is called for all changes to the router
        state. It is up to the router type driver to ignore state or state
        changes that it does not know or care about.
        """
        pass

    @abc.abstractmethod
    def update_router_postcommit(self, context, router_context):
        """Update a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the new
        state of the router, as well as the original state prior to the
        update_router call.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause the deletion of
        the resource.

        update_router_postcommit is called for all changes to the router
        state.  It is up to the routertype driver to ignore state or state
        changes that it does not know or care about.
        """
        pass

    @abc.abstractmethod
    def delete_router_precommit(self, context, router_context):
        """Perform operations specific to the routertype in
        preparation for the deletion of a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the current
        state of the router, prior to the call to delete it.

        Delete router resources previously allocated by this routertype
        driver for a router. Called inside transaction context on session.
        Runtime errors are not expected, but raising an exception will
        result in rollback of the transaction.
        """
        pass

    @abc.abstractmethod
    def delete_router_postcommit(self, context, router_context):
        """Delete a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the current
        state of the router, prior to the call to delete it.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Runtime errors are not expected, and will not
        prevent the resource from being deleted.
        """
        pass

    @abc.abstractmethod
    def schedule_router_precommit(self, context, router_context):
        """Perform operations specific to the routertype in preparation for
        scheduling of a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the current
        state of the router, prior to the call to schedule it.

        Perform operations that need to happen before scheduling of routers
        of this routertype. Called inside transaction context on session.
        Raising an exception will result in rollback of the transaction.
        """
        pass

    @abc.abstractmethod
    def schedule_router_postcommit(self, context, router_context):
        """Schedule the router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the current
        state of the router, prior to the call to schedule it.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause un-scheduling of
        the router.
        """
        pass

    @abc.abstractmethod
    def unschedule_router_precommit(self, context, router_context):
        """Perform operations specific to the routertype in preparation for
        un-scheduling of a router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the current
        state of the router, prior to the call to schedule it.

        Perform operations that need to happen before un-scheduling of routers
        of this routertype. Called inside transaction context on session.
        Raising an exception will result in rollback of the transaction.
        """
        pass

    @abc.abstractmethod
    def unschedule_router_postcommit(self, context, router_context):
        """Un-schedule the router.

        :param context: the neutron context of the request
        :param router_context: RouterContext instance describing the current
        state of the router, prior to the call to schedule it.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause re-scheduling of
        the router.
        """
        pass

    @abc.abstractmethod
    def add_router_interface_precommit(self, context, r_port_context):
        pass

    @abc.abstractmethod
    def add_router_interface_postcommit(self, context, r_port_context):
        pass

    @abc.abstractmethod
    def remove_router_interface_precommit(self, context, r_port_context):
        pass

    @abc.abstractmethod
    def remove_router_interface_postcommit(self, context, r_port_context):
        pass

    @abc.abstractmethod
    def create_floatingip_precommit(self, context, fip_context):
        """Create a floatingip.

        :param context: the neutron context of the request
        :param fip_context: FloatingipContext instance describing the new
        state of the floatingip

        Called before the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause the deletion of
        the resource.

        create_flotingip_precommit will not be used by most drivers. The
        only way a routertype driver can be known is to assume the default
        router type. This API was introduced to support allocation of
        floating IPs from NAT pools for Group Based Policy (GBP) workflow.
        """
        pass

    @abc.abstractmethod
    def create_floatingip_postcommit(self, context, fip_context):
        """Create a floatingip.

        :param context: the neutron context of the request
        :param fip_context: FloatingipContext instance describing the new
        state of the floatingip

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause the deletion of
        the resource.

        create_flotingip_postcommit is called for all changes to the router
        state.  It is up to the routertype driver to ignore state or state
        changes that it does not know or care about.
        """
        pass

    @abc.abstractmethod
    def update_floatingip_precommit(self, context, fip_context):
        """Perform operations specific to the routertype in preparation for
        the update of a floatingip.

        :param context: the neutron context of the request
        :param fip_context: FloatingipContext instance describing the new
        state of the floatingip, as well as the original state prior
        to the update_floatingip call.

        Update values of a floatingip, updating the associated resources
        in the database. Called inside transaction context on session.
        Raising an exception will result in rollback of the
        transaction.

        update_floatingip_precommit is called for all changes to the
        floatingip state. It is up to the routertype driver to ignore
        state or state changes that it does not know or care about.
        """
        pass

    @abc.abstractmethod
    def update_floatingip_postcommit(self, context, fip_context):
        """Update a floatingip.

        :param context: the neutron context of the request
        :param fip_context: FloatingipContext instance describing the new
        state of the floatingip, as well as the original state prior
        to the update_floatingip call.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Raising an exception will cause the deletion of
        the resource.

        update_flotingip_postcommit is called for all changes to the router
        state.  It is up to the routertype driver to ignore state or state
        changes that it does not know or care about.
        """
        pass

    @abc.abstractmethod
    def delete_floatingip_precommit(self, context, fip_context):
        """Perform operations specific to the routertype in preparation for
        the deletion of a floatingip.

        :param context: the neutron context of the request
        :param fip_context: FloatingipContext instance describing the current
        state of the floatingip, prior to the call to delete it.

        Delete floatingip resources previously allocated by this routertype
        driver for a floatingip. Called inside transaction context on session.
        Runtime errors are not expected, but raising an exception will result
        in rollback of the transaction.
        """
        pass

    @abc.abstractmethod
    def delete_floatingip_postcommit(self, context, fip_context):
        """Delete a floatingip.

        :param context: the neutron context of the request
        :param fip_context: FloatingipContext instance describing the current
        state of the floatingip, prior to the call to delete it.

        Called after the transaction commits. Call can block, though will
        block the entire process so care should be taken to not drastically
        affect performance. Runtime errors are not expected, and will not
        prevent the resource from being deleted.
        """
        pass

    # Methods for HA-enabled routers, all with default implementations
    def ha_interface_ip_address_needed(self, context, router, port,
                                       ha_settings_db, ha_group_uuid):
        """Determines if a router interface for an HA enabled router needs
        an extra IP address (in addition to the VIP address)

        :param context: the neutron context of the request
        :param router: dictionary of HA router
        :param port: dictionary of port associated with the HA group
        :param ha_settings_db: db object with ha settings
        :param ha_group_uuid: uuid of HA group's DB entry

        :returns - True if an extra IP address is needed, False otherwise.
        """
        import neutron_lib.constants as l3_constants
        if port['device_owner'] == l3_constants.DEVICE_OWNER_ROUTER_GW:
            return False
        else:
            return True
        return True

    def generate_ha_group_id(self, context, router, port, ha_settings_db,
                             ha_group_uuid):
        """Returns None or a unique integer value for use as identifier of an
        HSRP, VRRP or GLBP group configuration.

        :param context: the neutron context of the request
        :param router: dictionary of HA router
        :param port: dictionary of port associated with the HA group
        :param ha_settings_db: db object with ha settings
        :param ha_group_uuid: uuid of HA group's DB entry

        :returns - an integer value group identifier or None

        If None is returned, an identifier will be generated in a non-driver
        specific manner.
        """
        return None

    def get_ha_group_timers_parameters(self, context, router, port,
                                       ha_settings_db, ha_group_uuid):
        """Returns timers specific parameters for an HA group.

        :param context: the neutron context of the request
        :param router: dictionary of HA router
        :param port: dictionary of port associated with the HA group
        :param ha_settings_db: db object with ha settings
        :param ha_group_uuid: uuid of HA group's DB entry

        :returns - a text string with the parameters

        The text string is supposed to be understandable by the relevant
        config agent drivers corresponding to this router type driver.
        """
        return ""

    def get_ha_group_tracking_parameters(self, context, router, port,
                                         ha_settings_db, ha_group_uuid):
        """Returns object tracking parameters for an HA group.

        :param context: the neutron context of the request
        :param router: dictionary describing router
        :param port: dictionary of port associated with the HA group
        :param ha_settings_db: db object with ha settings
        :param ha_group_uuid: uuid of HA group's DB entry

        :returns - a text string with the parameters

        The text string is supposed to be understandable by the relevant
        config agent drivers corresponding to this router type driver.
        """
        return ""

    def get_other_ha_group_parameters(self, context, router, port,
                                      ha_settings_db, ha_group_uuid):
        """Returns arbitrary HA parameters for an HA group.

        :param context: the neutron context of the request
        :param router: dictionary describing router
        :param port: dictionary of port associated with the HA group
        :param ha_settings_db: db object with ha settings
        :param ha_group_uuid: uuid of HA group's DB entry

        :returns: a text string with the parameters

        The text string is supposed to be understandable by the relevant
        config agent drivers corresponding to this router type driver.
        """
        return ""

    def pre_backlog_processing(self, context):
        """Perform driver specific processing before backlog is processed.

        :param context: admin neutron context

        This function is invoked before the router plugin is processing the
        backlog of unscheduled routers. It allows the router type driver to
        perform any tasks that should be performed before the backlog is
        processed.
        """
        pass

    def post_backlog_processing(self, context):
        """Perform driver specific processing after backlog is processed.

        :param context: admin neutron context

        This function is invoked after the router plugin has processed the
        backlog of unscheduled routers. It allows the router type driver to
        perform any tasks that should be performed after the backlog has been
        processed.
        """
        pass
